import * as WebFont from 'webfontloader';
import * as PIXI from 'pixi.js-legacy';
import * as MedsurfDraw from '../lib';
import { v4 as uuidv4 } from 'uuid';
import * as Models from '@ascii-dev-user/medsurf-lib/models';
import { gsap } from 'gsap';
// @ts-ignore TODO ignore
import { PixiPlugin } from 'gsap/PixiPlugin.js';
// import { Simple, SpatialHash } from "pixi-cull"

/* TODO
    - Menu schliessen bei klonen / kopieren
    - Selftest visible when no selftest available
    Bugfixing machen
    +++
    - Pfeil / Spickel (Grundkante / Wedge 1/3)
    - Löschen von PositionPoints when no primitives / lines
    - Ersetzen des Generators im Tooltop -> polygon zu Fill generator
    - Generatoren austauschen -> alle als marker ??? oder nicht ???
    - Layergruppe verschieben -> move events einfügen
    - Rotation: convert to integer
    - Menu anzeigen -> Hauptmenu wird nicht immer korrekt ausgeblendet.
    - Toggle submenus -> change hitbox when open and reset selection when outside
    - Selektion der Fläche -> wird nicht geändert. scheint andere elemente zu ändern -> no events ???
    - Fill unset line color creates black lines -> fix this
    - Selftest add new element only for selftest -> use ellipse for this (make ellipse smaller once clicked) - otherwise elements like arrow / rectangle / ... will be shown (not nice)
    - FillLine: "isPointInPolygon" and "getNearestPointForPoint" function wrong;
    - Line: Hitbox not defined -> HOW TO ???
    - Selftest selcet interaction available -> WHY?
    - Speicherfehler
    - Menu verschiebung auf LayerGruppe when Schrift angepasst wird (How to???)
    Context menu schrift sowie anzeige nicht ok.
    Interactions anpassen; siehe dazu CloneInteraction -> kann einfacher so umgesetzt werden (Scale auf Draw nicht mehr notwendig; ausbauen) -> Test: Move events in Draw / Copy / Selection Generator
    Verpixelung ??? -> grössere Schrift
    Create Behaviour/Interaction -> to eliminate/clean up the elements
*/

// Element Refs
const canvasDiv = document.getElementById("medsurf_canvas_div") as HTMLDivElement;
const canvas = document.getElementById("medsurf_canvas") as HTMLCanvasElement;

const buttonDefault = document.getElementById("button_default") as HTMLButtonElement;
const buttonMarker = document.getElementById("button_marker") as HTMLButtonElement;
const buttonSelftest = document.getElementById("button_selftest") as HTMLButtonElement;
const buttonAuthor = document.getElementById("button_author") as HTMLButtonElement;
const buttonModel = document.getElementById("button_model") as HTMLButtonElement;
const buttonChange = document.getElementById("button_change") as HTMLButtonElement;
const buttonFilter = document.getElementById("button_filter") as HTMLButtonElement;
const buttonPosition = document.getElementById("button_position") as HTMLButtonElement;
const buttonVirtualPointer = document.getElementById("button_virtual-poiner") as HTMLButtonElement;
const buttonTest = document.getElementById("button_test") as HTMLButtonElement;
const buttonTest2 = document.getElementById("button_test2") as HTMLButtonElement;
const buttonTest3 = document.getElementById("button_test3") as HTMLButtonElement;

// PIXI SETTINGS
PIXI.settings.SCALE_MODE = PIXI.SCALE_MODES.NEAREST;

/**
 * Ready function
 */
function ready(fn: any) {
    setTimeout(() => {
        // see if DOM is already available
        if (document.readyState === "complete" || document.readyState === "interactive") {
            // call on next available tick
            setTimeout(fn, 1);
        } else {
            document.addEventListener("DOMContentLoaded", fn);
        }
    }, 1000);
}

/**
 * Defaults
 */
let app: PIXI.Application;
let loader: PIXI.Loader;
let virtualPointer: MedsurfDraw.VirtualPointerElement;
let image: MedsurfDraw.Image | MedsurfDraw.DeepZoomImage;
const layerGroupsArray: {[key: string]: Models.LayerGroup; } = {};
const elementsArray: {[key: string]: MedsurfDraw.ImageObject; } = {};

/**
 * Destroy image
 */
async function destroyImage() {
    return new Promise<void>(resolve => {
        gsap.to(image, {
            duration: 0.3,
            alpha: 0,
            onComplete: () => {
                // Remove sprite from stage
                app.stage.removeChild(image);
                // Events
                image.removeAllListeners();
                // Destroy sprite
                image.destroy({children: true});

                // Resolve
                resolve();
            }
        });
    });
}

/**
 * Setup image
 */
async function setupImage() {
    //<editor-fold desc="Images">
    // Image
    /**/
    image = new MedsurfDraw.Image({
        renderer: app.renderer,
        canvas: canvas,
        loader: loader,
        imageId: 'imageId',
        id: "3893",
        /* safe * /
        endpointUrl: "https://api.medsurf.stage.iml.unibe.ch/",
        dimensions: {width: 1500, height: 1514},
        fullSize: "/slides/3893/image/1500x1514_80_eaaa24fd-e69e-491e-b576-fccab29c85a2.jpg"
        /* lama */
        // url: "lama.jpeg",
        url: "https://s3-zh.os.switch.ch/medsurf/public/slide/2701,2701/image/1800x1800_80.jpeg?c9eab3bbb580aabcd73c1f7f08fb1a8b",
        dimensions: {width: 1152, height: 864},
        /** /
         /* best * /
         endpointUrl: "",
         dimensions: {width: 720, height: 540},
         fullSize: "a/249309_2261068168664_7462534_n(2).jpg"
         /**/
    });
    /**/
    /** /
     // Deep zoom image
     let image = new MedsurfDraw.DeepZoomImage({
                    renderer: app.renderer,
                    canvas: canvas,
                    loader: loader,
                    id: "5901",
                    dimensions: { width: 20944, height: 23965, tileSize: 256},
                    endpointUrl: "https://api.medsurf.iml.unibe.ch/",
                    fullSize: "",
                });
     /**/

    // Add sprite to stage
    app.stage.addChild(image);
    //</editor-fold>

    //<editor-fold desc="Image Events">
    image.on('sendError', (e) => console.log('sendError', e));
    // Grouping
    image.on('editGrouping', (g: Models.Grouping) => console.log('editGrouping', g));
    image.on('editLayerGroup', (lg: Models.LayerGroup) => console.log('editLayerGroup', lg));
    image.on('createArrowWithInteractiveGrouping', () => console.log('createArrowWithInteractiveGrouping'));
    image.on('createArrowWithPoiGrouping', () => console.log('createArrowWithPoiGrouping'));
    image.on('createArrowWithTextGrouping', () => console.log('createArrowWithTextGrouping'));
    image.on('createEllipseWithInteractiveGrouping', () => console.log('createEllipseWithInteractiveGrouping'));
    image.on('createEllipseWithPoiGrouping', () => console.log('createEllipseWithPoiGrouping'));
    image.on('createEllipseWithTextGrouping', () => console.log('createEllipseWithTextGrouping'));
    image.on('createRectangleWithInteractiveGrouping', () => console.log('createRectangleWithInteractiveGrouping'));
    image.on('createRectangleWithPoiGrouping', () => console.log('createRectangleWithPoiGrouping'));
    image.on('createRectangleWithTextGrouping', () => console.log('createRectangleWithTextGrouping'));
    image.on('createInteractiveGrouping', (ios: Models.ImageObject[]) => console.log('createInteractiveGrouping', ios));
    image.on('createPoiGrouping', (ios: Models.ImageObject[]) => console.log('createPoiGrouping', ios));
    image.on('createTextGrouping', (ios: Models.ImageObject[]) => console.log('createTextGrouping', ios));
    image.on('createLegendGrouping', (ios: Models.ImageObject[]) => console.log('createLegendGrouping', ios));
    // Legend
    image.on('createLegend', () => console.log('createLegend'));
    // -- LayerGroup Events
    image.on('unselectGrouping', () => console.log('unselectGrouping'));
    image.on('selectLayerGroup', (lg: Models.LayerGroup) => console.log('selectLayerGroup', lg));
    image.on('selectLayerGroupByImageObject', (io: Models.ImageObject) => console.log('selectLayerGroupByImageObject', io));
    image.on('unselectLayerGroup', () => console.log('unselectLayerGroup'));
    image.on('addLayerGroup', (lg: Models.LayerGroup) => console.log('addLayerGroup', lg));
    image.on('addLayerGroupByName', (lgn: string) => console.log('addLayerGroupByName', lgn));
    image.on('copyLayerGroup', (lg: Models.LayerGroup) => {
        const generator = new MedsurfDraw.CloneLayerGenerator({
            target: image,
            imageObjects: image.getImageObjects().map(io => (io as any).model),
            cloneLayerGroup: lg,
            layerGroupMappings: [],
            imageObjectMappings: [],
            // TODO layerGroup: sketch,
            image: {}
        });
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('clipboardLayerGroup', (lg: Models.LayerGroup) => console.log('clipboardLayerGroup', lg));
    image.on('updateLayerGroup', (lg: Models.LayerGroup) => console.log('updateLayerGroup', lg));
    image.on('deleteLayerGroup', (lg: Models.LayerGroup) => console.log('deleteLayerGroup', lg));
    image.on('addLayerGroupImageObjectByImageObject', (io: Models.ImageObject) => console.log('addLayerGroupImageObjectByImageObject', io));
    image.on('deleteLayerGroupImageObjectByImageObject', (io: Models.ImageObject) => console.log('deleteLayerGroupImageObjectByImageObject', io));
    // ImageObject Events
    image.on('editImageObject', (io: Models.ImageObject) => console.log('editImageObject', io));
    image.on('selectImageObject', (io: Models.ImageObject) => console.log('selectImageObject', io));
    image.on('unselectImageObject', () => console.log('unselectImageObject'));
    image.on('addImageObject', (io: Models.ImageObject, layer: Models.LayerGroup, image: Models.Image) => {
        console.log("addImageObject", io);
        if (!layer.layerImageObjects) {
            layer.layerImageObjects = [];
        }
        if (layer.layerImageObjects.findIndex(lio => lio.imageObject === io.id) === -1) {
            const layerImageObject: Models.LayerImageObject = {
                dirty: true,
                id: uuidv4(),
                order: layer.layerImageObjects.length,
                layerGroup: layerGroupsArray['sketch'],
                imageObject: io.id as string
            };
            layer.layerImageObjects.push(layerImageObject);
        }
    });
    image.on('copyImageObject', (io: Models.PositionPoint | Models.Line | Models.FillCollection | Models.LegendCollection) => {
        const generator = new MedsurfDraw.CloneItemGenerator({
            target: image,
            imageObjects: image.getImageObjects().map(io => (io as any).model),
            element: io,
            imageObjectMappings: [],
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('clipboardImageObject', (io: Models.ImageObject) => console.log('clipboardImageObject', io));
    image.on('updateImageObject', (io: Models.ImageObject) => console.log('updateImageObject', io));
    image.on('degroupImageObjectFromLayerGroup', (layerGroup: Models.LayerGroup, io: Models.ImageObject) => console.log('degroupImageObjectFromLayerGroup', layerGroup, io));
    image.on('degroupImageObjectFromGrouping', (grouping: Models.Grouping, io: Models.ImageObject) => console.log('degroupImageObjectFromGrouping', grouping, io));
    image.on('deleteImageObject', (io: Models.ImageObject) => {
        console.log("deleteImageObject", io);
        if (!layerGroupsArray['sketch'].layerImageObjects) {
            layerGroupsArray['sketch'].layerImageObjects = [];
        }
        const layerImageObjectIndex = layerGroupsArray['sketch'].layerImageObjects.findIndex(lio => lio.imageObject === io.id);
        if (layerImageObjectIndex > -1) {
            layerGroupsArray['sketch'].layerImageObjects.splice(layerImageObjectIndex, 1);
        }
    });
    // -- Primitive Events
    image.on('editPrimitive', (p: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive
                                   | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive,
                               pp: Models.PositionPoint) => console.log('editPrimitive', p, pp));
    image.on('selectPrimitive', (p: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive
                                     | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive,
                                 pp: Models.PositionPoint) => console.log('selectPrimitive', p, pp));
    image.on('unselectPrimitive', () => console.log('unselectPrimitive'));
    image.on('updatePrimitive', (p: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive
        | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive) => console.log('updatePrimitive', p));
    image.on('deletePrimitive', (p: Models.ArrowPrimitive | Models.BackgroundRectanglePrimitive
        | Models.EllipsePrimitive | Models.RectanglePrimitive | Models.TextPrimitive) => console.log('deletePrimitive', p));
    // -- Legend column events
    image.on('selectLegendColumn', (lc: Models.LegendColumn, l: Models.LegendCollection) => console.log('selectLegendColumn', lc, l));
    image.on('unselectLegendColumn', () => console.log('unselectLegendColumn'));
    image.on('updateLegendColumn', (lc: Models.LegendColumn) => console.log('updateLegendColumn', lc));
    image.on('deleteLegendColumn', (lc: Models.LegendColumn) => console.log('deleteLegendColumn', lc));
    image.on('editLegendColumn', (lc: Models.LegendColumn) => console.log('editPrimitive', lc));
    // -- Legend row events
    image.on('selectLegendRow', (lr: Models.LegendRow, lc: Models.LegendColumn, l: Models.LegendCollection) => console.log('selectLegendRow', lr, lc, l));
    image.on('unselectLegendRow', () => console.log('unselectLegendRow'));
    image.on('updateLegendRow', (lr: Models.LegendRow) => console.log('updateLegendRow', lr));
    image.on('deleteLegendRow', (lr: Models.LegendRow) => console.log('deleteLegendRow', lr));
    image.on('editLegendRow', (lr: Models.LegendRow) => console.log('editLegendRow', lr));
    // -- Generator events
    image.on('createPolygon', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.PolygonGenerator({
            target: image,
            // bezierCourve: false,
            stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
            // startElement: test17Point,
            layerGroup: layerGroupsArray['layerGroupModel1'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createFill', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.FillGenerator({
            target: image,
            // bezierCourve: false,
            stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
            // startElement: test17Point,
            layerGroup: layerGroupsArray['layerGroupModel1'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createInteractiveArea', () => {
        setTimeout(() => {
            const generator = new MedsurfDraw.InteractiveAreaGenerator({
                target: image,
                bezierCourve: false,
                stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                // startElement: test17Point,
                layerGroup: layerGroupsArray['sketch'],
                image: {}
            });
            // @ts-ignore
            generator.once('startGenerator', _startGenerator, this);
            generator.once('endGenerator', generator.endGenerator, generator);
            generator.once('abortGenerator', generator.abortGenerator, generator);
        }, 200);
    });
    image.on('createText', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.TextPrimitiveGenerator({
            target: image,
            stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
            // startElement: test17Point,
            layerGroup: layerGroupsArray['sketch'],
            image: {},
            rotateWithLine: false,
            withScaling: false,
            triggerImmediateEndDraw: true
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createArrow', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.ArrowGenerator({
            target: image,
            layerGroup: layerGroupsArray['sketch'],
            image: {},
            hasWedgePoint: true
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createArrowWithPoiGrouping', () => {
        // TODO
        throw 'not implemented';
    });
    image.on('createEllipse', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.EllipseGenerator({
            target: image,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createEllipseWithPoiGrouping', () => {
        // TODO
        throw 'not implemented';
    });
    image.on('createRectangle', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.RectangleGenerator({
            target: image,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createRectangleWithPoiGrouping', () => {
        // TODO
        throw 'not implemented';
    });
    image.on('createArrowMarker', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.PrimitiveMarkerGenerator({
            target: image,
            type: Models.ImageObjectType.ARROW,
            hasWedgePoint: false,
            rotateWithLine: true,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        } as MedsurfDraw.ArrowMarkerGeneratorModel);
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createEllipseMarker', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.PrimitiveMarkerGenerator({
            target: image,
            type: Models.ImageObjectType.ELLIPSE,
            rotateWithLine: true,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('createRectangleMarker', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.PrimitiveMarkerGenerator({
            target: image,
            type: Models.ImageObjectType.RECTANGLE,
            rotateWithLine: true,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
        // @ts-ignore
        generator.once('startGenerator', _startGenerator, this);
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    image.on('cloneFromClipboard', () => {
        image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
        const generator = new MedsurfDraw.CloneLayerGenerator({
            target: image,
            imageObjects: image.getImageObjects().map((io) => (io as any).model),
            cloneLayerGroup: layerGroupsArray['layerGroupModel1'],
            imageObjectMappings: [],
            layerGroupMappings: [],
            // TODO layerGroup: sketch,
            image: {},
        });
        generator.once('endGenerator', generator.endGenerator, generator);
        generator.once('abortGenerator', generator.abortGenerator, generator);
    });
    //</editor-fold>

    //<editor-fold desc="Debug">
    // @ts-ignore
    window["sprite"] = image;
    //</editor-fold>

    // Start app
    app.stage.emit("setDefaultMode", "default");
    app.stage.interactive = true;

    return Promise.resolve();
}

/**
 * Setup image objects
 */
async function setupImageObjects() {
    //<editor-fold desc="Layer groups">
    // Sketch
    layerGroupsArray['sketch'] = {
        id: 'group1',
        image: image,
        order: 0,
        name: 'Sketch',
        groupContents: [],
        layerImageObjects: []
    };

    // ---- Sketch group content 1
    const sketchGroupContent1: Models.LayerGroupContent = {
        id: 'sketchGroupContent1',
        content: {
            title: "test",
            body: "blab"
        },
        order: 0,
    }
    // @ts-ignore
    layerGroupsArray['sketch'].groupContents.push(sketchGroupContent1);

    // ---- Sketch image object 1
    const sketchImageObject1: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['sketch'],
        order: 1,
        imageObject: "ellipsePoint",
    };
    // @ts-ignore
    layerGroupsArray['sketch'].layerImageObjects.push(sketchImageObject1);

    // Add to image
    image.layerGroups.push(layerGroupsArray['sketch']);

    // --------------------------------------
    // -- Layer group 1
    layerGroupsArray['layerGroupModel1'] = {
        id: 'group2',
        image: image,
        order: 0,
        name: 'test',
        groupContents: [],
        layerImageObjects: [],
    };

    // ---- Layer group 1 image object 1
    const layerImageObject1: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['layerGroupModel1'],
        order: 1,
        imageObject: "arrowPoint",
    };
    // @ts-ignore
    layerGroupsArray['layerGroupModel1'].layerImageObjects.push(layerImageObject1);

    // ---- Layer group 1 image object 2
    const layerImageObject2: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['layerGroupModel1'],
        order: 2,
        imageObject: "rectanglePoint",
    };
    // @ts-ignore
    layerGroupsArray['layerGroupModel1'].layerImageObjects.push(layerImageObject2);

    // ---- Layer group 1 image object 3
    const layerImageObject3: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['layerGroupModel1'],
        order: 3,
        imageObject: "line1",
    };
    // @ts-ignore
    layerGroupsArray['layerGroupModel1'].layerImageObjects.push(layerImageObject3);

    // Add to image
    image.layerGroups.push(layerGroupsArray["layerGroupModel1"]);

    // --------------------------------------
    // -- Real Layer Group 1
    layerGroupsArray['realLayerGroup1'] = {
        id: 'realLayerGroup1',
        image: image,
        order: 0,
        name: 'Blubber',
        groupContents: [],
        layerImageObjects: [],
    };

    // ---- Real Layer Image Object 1
    const realLayerImageObject1: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['realLayerGroup1'],
        order: 1,
        imageObject: "realPoint1",
    };
    // @ts-ignore
    layerGroupsArray['realLayerGroup1'].layerImageObjects.push(realLayerImageObject1);

    // ---- Real Layer Image Object 2
    const realLayerImageObject2: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['realLayerGroup1'],
        order: 2,
        imageObject: "realPoint2",
    };
    // @ts-ignore
    layerGroupsArray['realLayerGroup1'].layerImageObjects.push(realLayerImageObject2);

    // ---- Real Layer Image Object 3
    const realLayerImageObject3: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['realLayerGroup1'],
        order: 3,
        imageObject: "realPoint3",
    };
    // @ts-ignore
    layerGroupsArray['realLayerGroup1'].layerImageObjects.push(realLayerImageObject3);

    // ---- Real Layer Image Object 4
    const realLayerImageObject4: Models.LayerImageObject = {
        layerGroup: layerGroupsArray['realLayerGroup1'],
        order: 4,
        imageObject: "realLine1",
    };
    // @ts-ignore
    layerGroupsArray['realLayerGroup1'].layerImageObjects.push(realLayerImageObject4);

    // Add to image
    image.layerGroups.push(layerGroupsArray["realLayerGroup1"]);
    //</editor-fold>

    //<editor-fold desc="Elements">
    //<editor-fold desc="Tests">
    /**/
    // -- Arrow point
    const arrowPoint = new MedsurfDraw.PositionPoint(image, {
        id: 'arrowPoint',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 100 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint', 'line1', 'textPoint'],
        primitives: []
    });
    image.addChild(arrowPoint);
    // -- Rectangle point
    const rectanglePoint = new MedsurfDraw.PositionPoint(image, {
        id: 'rectanglePoint',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 200 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['arrowPoint'],
        primitives: []
    });
    image.addChild(rectanglePoint);
    // -- Ellipse point
    const ellipsePoint = new MedsurfDraw.PositionPoint(image, {
        id: 'ellipsePoint',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 100 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint', 'legendText2_2'],
        primitives: []
    });
    image.addChild(ellipsePoint);
    // -- Text point
    const textPoint = new MedsurfDraw.PositionPoint(image, {
        id: 'textPoint',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 200 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint'],
        primitives: []
    });
    image.addChild(textPoint);
    elementsArray['textPoint'] = textPoint;
    // -- Test point
    const testPoint = new MedsurfDraw.PositionPoint(image, {
        id: 'testPoint',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 300 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint'],
        primitives: []
    });
    image.addChild(testPoint);
    // -- Test 2 point
    const test2Point = new MedsurfDraw.PositionPoint(image, {
        id: 'test2Point',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 300 / image.dimensions.width,
            y: 300 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint'],
        primitives: []
    });
    image.addChild(test2Point);
    // -- Test 3 point
    const test3Point = new MedsurfDraw.PositionPoint(image, {
        id: 'test3Point',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 150 / image.dimensions.width,
            y: 150 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint'],
        primitives: []
    });
    image.addChild(test3Point);
    // -- Test point
    const test17Point = new MedsurfDraw.PositionPoint(image, {
        id: 'test17Point',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 50 / image.dimensions.width,
            y: 50 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: ['rectanglePoint'],
        primitives: []
    });
    image.addChild(test17Point);
    /**/
    /**/
    // -- Line1
    const line1 = new MedsurfDraw.Line(image, {
        id: 'line1',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'arrowPoint',
        end: 'rectanglePoint',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0xFF0000,
            hasFill: false,
        }
    });
    image.addChild(line1);
    elementsArray['line1'] = line1;
    // -- Line2
    const line2 = new MedsurfDraw.Line(image, {
        id: 'line2',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'ellipsePoint',
        end: 'arrowPoint',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0x0000FF,
            hasFill: false,
        }
    });
    image.addChild(line2);
    elementsArray['line2'] = line2;
    // -- Line3
    const line3 = new MedsurfDraw.Line(image, {
        id: 'line3',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'textPoint',
        end: 'ellipsePoint',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0x00FF00,
            hasFill: false,
        }
    });
    image.addChild(line3);
    elementsArray['line3'] = line3;
    // -- Line4
    const line4 = new MedsurfDraw.Line(image, {
        id: 'line4',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'rectanglePoint',
        end: 'textPoint',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0xFFFF00,
            hasFill: false,
        }
    });
    image.addChild(line4);
    elementsArray['line4'] = line4;
    // -- Line5
    const line5 = new MedsurfDraw.Line(image, {
        id: 'line5',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'rectanglePoint',
        end: 'testPoint',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0x00FFFF,
            hasFill: false,
        }
    });
    image.addChild(line5);
    elementsArray['line5'] = line5;
    // -- Arrow
    const arrow = new MedsurfDraw.ArrowPrimitive(image, {
        id: 'arrow',
        type: Models.ImageObjectType.ARROW,
        isStatic: false,
        zIndex: 0,
        positionPoint: arrowPoint.model,
        rotateWithLine: true,
        rectangle: new PIXI.Rectangle(
            0,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        dashFill: 2,
        dashEmpty: 2,
        options: {
            hasLine: true,
            lineColor: 0x00FF00,
            hasFill: true,
            fillColor: 0xFF0000,
            fillAlpha: 0.5,
        }
    });
    // TODO this -> really needed
    // @ts-ignore
    arrowPoint.model.primitives.push(arrow.model);
    arrowPoint.addChild(arrow);
    // -- Rectangle
    const rect = new MedsurfDraw.RectanglePrimitive(image, {
        id: 'rect',
        type: Models.ImageObjectType.RECTANGLE,
        isStatic: false,
        zIndex: 0,
        positionPoint: rectanglePoint.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            50 / image.dimensions.width,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        },
        angle: 45 * Math.PI / 180,
    });
    // TODO this -> really needed
    // @ts-ignore
    rectanglePoint.model.primitives.push(rect.model);
    rectanglePoint.addChild(rect);
    // -- Ellipse
    const ellipse = new MedsurfDraw.EllipsePrimitive(image, {
        id: 'ellipse',
        type: Models.ImageObjectType.ELLIPSE,
        isStatic: false,
        zIndex: 0,
        positionPoint: ellipsePoint.model,
        rotateWithLine: false,
        ellipse: new PIXI.Ellipse(
            0,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // TODO this -> really needed
    // @ts-ignore
    ellipsePoint.model.primitives.push(ellipse.model);
    ellipsePoint.addChild(ellipse);
    // -- Text
    const text = new MedsurfDraw.TextPrimitive(image, {
        id: 'text',
        type: Models.ImageObjectType.TEXT,
        isStatic: false,
        zIndex: 0,
        positionPoint: textPoint.model,
        rotateWithLine: false,
        text: "Hello Text",
        rectangle: {
            x: 60 / image.dimensions.width,
            y: 0,
            width: 0,
            height: 0,
        },
        style: {
            fill: 0xe2d62d,
            fontSize: 32 / image.dimensions.width,
            fontFamily: "proxima_nova_altsemibold",
            align: 'left',
            breakWords: false,
        }
    });
    // TODO this -> really needed
    // @ts-ignore
    textPoint.model.primitives.push(text.model);
    textPoint.addChild(text);
    elementsArray['text'] = text;
    // -- Background Rectangle
    const backgroundRectangle = new MedsurfDraw.BackgroundRectanglePrimitive(image, {
        id: 'backgroundRectangle',
        type: Models.ImageObjectType.BACKGROUNDRECTANGLE,
        isStatic: false,
        zIndex: 0,
        positionPoint: textPoint.model,
        rotateWithLine: false,
        primitiveId: 'text',
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FFFF,
            fillAlpha: 0.6
        },
    });
    // TODO this -> really needed
    // @ts-ignore
    textPoint.model.primitives.push(backgroundRectangle.model);
    textPoint.addChild(backgroundRectangle);
    elementsArray['backgroundRectangle'] = backgroundRectangle;
    // -- Fill
    const fill1 = new MedsurfDraw.FillCollection(image, {
        id: 'fill1',
        type: Models.ImageObjectType.FILLCOLLECTION,
        isStatic: false,
        zIndex: 0,
        isInteractive: true,
        lines: [line1.model.id, line2.model.id, line3.model.id, line4.model.id] as string[],
        interactiveItems: ['rectanglePoint', 'legendText2_2'],
        options: {
            hasLine: false,
            hasFill: true,
            fillColor: 0xFF00FF,
            fillAlpha: 0.35
        }
    });
    image.addChild(fill1);
    elementsArray['fill1'] = fill1;
    // -- FillLine
    const fillLine = new MedsurfDraw.Line(image, {
        id: 'fillLine',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'test2Point',
        end: 'fill1',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: false,
        options: {
            hasLine: true,
            lineColor: 0xFFFFFF,
            hasFill: false,
        }
    });
    image.addChild(fillLine);
    elementsArray['fillLine'] = fillLine;
    // -- FillLine2
    const fillLine2 = new MedsurfDraw.Line(image, {
        id: 'fillLine2',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'test3Point',
        end: 'fill1',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: false,
        options: {
            hasLine: true,
            lineColor: 0xFFFFFF,
            hasFill: false,
        }
    });
    image.addChild(fillLine2);
    elementsArray['fillLine2'] = fillLine2;
    // -- Legend model
    const legendCollectionModel: Models.LegendCollection = {
        id: 'legend',
        type: Models.ImageObjectType.LEGENDCOLLECTION,
        isStatic: false,
        zIndex: 0,
        rectangle: new PIXI.Rectangle(
            50 / image.dimensions.width,
            400 / image.dimensions.width,
            400 / image.dimensions.width,
            100 / image.dimensions.width
        ),
        cols: [],
        rowDistance: 14 / image.dimensions.width,
        columnDistance: 14 / image.dimensions.width,
        indexWidth: 14 / image.dimensions.width
    };
    // ---- Legend column 1 model
    const legendColumn1Model: Models.LegendColumn = {
        id: 'legendColumn1',
        type: Models.ImageObjectType.LEGENDCOLUMN,
        isStatic: false,
        zIndex: 0,
        table: legendCollectionModel,
        rows: []
    };
    legendCollectionModel.cols.push(legendColumn1Model);
    // ------ Legend row 1_1 model
    const legendText1_1Model: Models.LegendRow = {
        id: 'legendText1_1',
        type: Models.ImageObjectType.LEGENDROW,
        isStatic: false,
        zIndex: 0,
        text: 'row1\nrow1',
        style: {
            fill: 0xFF00FF,
            fontSize: 32 / image.dimensions.width,
            fontFamily: 'proxima_nova_altsemibold',
            align: 'left',
            breakWords: false,
        },
        col: legendColumn1Model
    };
    legendColumn1Model.rows.push(legendText1_1Model);
    // ------ Legend row 1_2 model
    const legendText1_2Model: Models.LegendRow = {
        id: 'legendText1_2',
        type: Models.ImageObjectType.LEGENDROW,
        isStatic: false,
        zIndex: 0,
        text: 'row1',
        style: {
            fill: 0x0000FF,
            fontSize: 32 / image.dimensions.width,
            fontFamily: 'proxima_nova_altsemibold',
            align: 'left',
            breakWords: false,
        },
        col: legendColumn1Model
    };
    legendColumn1Model.rows.push(legendText1_2Model);
    // ---- Legend column 2 model
    const legendColumn2Model: Models.LegendColumn = {
        id: 'legendColumn2',
        type: Models.ImageObjectType.LEGENDCOLUMN,
        isStatic: false,
        zIndex: 0,
        table: legendCollectionModel,
        rows: []
    };
    legendCollectionModel.cols.push(legendColumn2Model);
    // ------ Legend row 2_1 model
    const legendText2_1Model: Models.LegendRow = {
        id: 'legendText2_1',
        type: Models.ImageObjectType.LEGENDROW,
        isStatic: false,
        zIndex: 0,
        text: 'row2',
        style: {
            fill: 0xFF00FF,
            fontSize: 32 / image.dimensions.width,
            fontFamily: 'proxima_nova_altsemibold',
            align: 'left',
            breakWords: false,
        },
        col: legendColumn2Model
    };
    legendColumn2Model.rows.push(legendText2_1Model);
    // ------ Legend row 2 model
    const legendText2_2Model: Models.LegendRow = {
        id: 'legendText2_2',
        type: Models.ImageObjectType.LEGENDROW,
        isStatic: false,
        zIndex: 0,
        text: 'row2',
        style: {
            fill: 0x0000FF,
            fontSize: 32 / image.dimensions.width,
            fontFamily: 'proxima_nova_altsemibold',
            align: 'left',
            breakWords: false,
        },
        col: legendColumn2Model
    };
    legendColumn2Model.rows.push(legendText2_2Model);
    // Legend
    const legend = new MedsurfDraw.LegendCollection(image, legendCollectionModel);
    image.addChild(legend);
    elementsArray['legend'] = legend;
    /**/
    //</editor-fold>
    // TEST LINE
    //<editor-fold desc="Real Test">
    // -- Real Point 1
    const realPoint1 = new MedsurfDraw.PositionPoint(image, {
        id: 'realPoint1',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 500 / image.dimensions.width,
            y: 500 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(realPoint1);
    // -- Real Point 2
    const realPoint2 = new MedsurfDraw.PositionPoint(image, {
        id: 'realPoint2',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 600 / image.dimensions.width,
            y: 500 / image.dimensions.width,
        },
        isSelftest: false,
        selftestItems: [],
        primitives: []
    });
    image.addChild(realPoint2);
    // -- Real Point 3
    const realPoint3 = new MedsurfDraw.PositionPoint(image, {
        id: 'realPoint3',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 550 / image.dimensions.width,
            y: 550 / image.dimensions.width,
        },
        isSelftest: false,
        selftestItems: [],
        primitives: []
    });
    image.addChild(realPoint3);
    // -- Real Line 1
    const realLine1 = new MedsurfDraw.Line(image, {
        id: 'realLine1',
        type: Models.ImageObjectType.LINE,
        isStatic: false,
        zIndex: 0,
        start: 'realPoint1',
        end: 'realPoint2',
        isBezier: true,
        strokeWidth: 1 / image.dimensions.width,
        interactWithLines: true,
        options: {
            hasLine: true,
            lineColor: 0xFF0000,
            hasFill: false,
        }
    });
    image.addChild(realLine1);
    elementsArray['realLine1'] = realLine1;
    // -- Real Ellipse 1
    const realEllipse1 = new MedsurfDraw.EllipsePrimitive(image, {
        id: 'realEllipse1',
        type: Models.ImageObjectType.ELLIPSE,
        isStatic: false,
        zIndex: 0,
        positionPoint: realPoint1.model,
        rotateWithLine: false,
        ellipse: new PIXI.Ellipse(
            0,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    realPoint1.model.primitives.push(realEllipse1.model);
    realPoint1.addChild(realEllipse1);
    // -- Real Text 1
    const realText1 = new MedsurfDraw.TextPrimitive(image, {
        id: 'realText1',
        type: Models.ImageObjectType.TEXT,
        isStatic: false,
        zIndex: 0,
        positionPoint: realPoint2.model,
        rotateWithLine: false,
        text: "Hello Text",
        rectangle: {
            x: 60 / image.dimensions.width,
            y: 0,
            width: 0,
            height: 0,
        },
        style: {
            fill: 0xe2d62d,
            fontSize: 32 / image.dimensions.width,
            fontFamily: "proxima_nova_altsemibold",
            align: 'left',
            breakWords: false,
        }
    });
    // @ts-ignore
    realPoint2.model.primitives.push(realText1.model);
    realPoint2.addChild(realText1);
    //</editor-fold>
    // TEST Primitives
    //<editor-fold desc="Scale Test">
    //<editor-fold desc="Ellipse Primitive">
    //<editor-fold desc="Test 1">
    // -- ellipse Primitive Test Point 1
    const ellipsePrimitiveTestPoint1 = new MedsurfDraw.PositionPoint(image, {
        id: 'ellipsePrimitiveTestPoint1',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 800 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(ellipsePrimitiveTestPoint1);
    // -- ellipse Primitive Test 1
    const ellipsePrimitiveTest1 = new MedsurfDraw.EllipsePrimitive(image, {
        id: 'ellipsePrimitiveTest1',
        type: Models.ImageObjectType.ELLIPSE,
        isStatic: false,
        zIndex: 0,
        positionPoint: ellipsePrimitiveTestPoint1.model,
        rotateWithLine: false,
        ellipse: new PIXI.Ellipse(
            0,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    ellipsePrimitiveTestPoint1.model.primitives.push(ellipsePrimitiveTest1.model);
    ellipsePrimitiveTestPoint1.addChild(ellipsePrimitiveTest1);
    //</editor-fold>
    //<editor-fold desc="Test 2">
    // -- ellipse Primitive Test Point 2
    const ellipsePrimitiveTestPoint2 = new MedsurfDraw.PositionPoint(image, {
        id: 'ellipsePrimitiveTestPoint2',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 800 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(ellipsePrimitiveTestPoint2);
    // -- ellipse Primitive Test 2
    const ellipsePrimitiveTest2 = new MedsurfDraw.EllipsePrimitive(image, {
        id: 'ellipsePrimitiveTest2',
        type: Models.ImageObjectType.ELLIPSE,
        isStatic: false,
        zIndex: 0,
        positionPoint: ellipsePrimitiveTestPoint2.model,
        rotateWithLine: false,
        ellipse: new PIXI.Ellipse(
            0,
            0,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        angle: 0.7853982,
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    ellipsePrimitiveTestPoint2.model.primitives.push(ellipsePrimitiveTest2.model);
    ellipsePrimitiveTestPoint2.addChild(ellipsePrimitiveTest2);
    //</editor-fold>
    //<editor-fold desc="Test 3">
    // -- ellipse Primitive Test Point 3
    const ellipsePrimitiveTestPoint3 = new MedsurfDraw.PositionPoint(image, {
        id: 'ellipsePrimitiveTestPoint3',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 800 / image.dimensions.width,
            y: 300 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(ellipsePrimitiveTestPoint3);
    // -- ellipse Primitive Test 3
    const ellipsePrimitiveTest3 = new MedsurfDraw.EllipsePrimitive(image, {
        id: 'ellipsePrimitiveTest3',
        type: Models.ImageObjectType.ELLIPSE,
        isStatic: false,
        zIndex: 0,
        positionPoint: ellipsePrimitiveTestPoint3.model,
        rotateWithLine: false,
        ellipse: new PIXI.Ellipse(
            10 / image.dimensions.width,
            10 / image.dimensions.width,
            20 / image.dimensions.width,
            20 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    ellipsePrimitiveTestPoint3.model.primitives.push(ellipsePrimitiveTest3.model);
    ellipsePrimitiveTestPoint3.addChild(ellipsePrimitiveTest3);
    //</editor-fold>
    //</editor-fold>
    //<editor-fold desc="Rectangle Primitive">
    //<editor-fold desc="Test 1">
    // -- rectangle Primitive Test Point 1
    const rectanglePrimitiveTestPoint1 = new MedsurfDraw.PositionPoint(image, {
        id: 'rectanglePrimitiveTestPoint1',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 900 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(rectanglePrimitiveTestPoint1);
    // -- rectangle Primitive Test 1
    const rectanglePrimitiveTest1 = new MedsurfDraw.RectanglePrimitive(image, {
        id: 'rectanglePrimitiveTest1',
        type: Models.ImageObjectType.RECTANGLE,
        isStatic: false,
        zIndex: 0,
        positionPoint: rectanglePrimitiveTestPoint1.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            0,
            0,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    rectanglePrimitiveTestPoint1.model.primitives.push(rectanglePrimitiveTest1.model);
    rectanglePrimitiveTestPoint1.addChild(rectanglePrimitiveTest1);
    //</editor-fold>
    //<editor-fold desc="Test 2">
    // -- rectangle Primitive Test Point 2
    const rectanglePrimitiveTestPoint2 = new MedsurfDraw.PositionPoint(image, {
        id: 'rectanglePrimitiveTestPoint2',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 900 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(rectanglePrimitiveTestPoint2);
    // -- rectangle Primitive Test 2
    const rectanglePrimitiveTest2 = new MedsurfDraw.RectanglePrimitive(image, {
        id: 'rectanglePrimitiveTest2',
        type: Models.ImageObjectType.RECTANGLE,
        isStatic: false,
        zIndex: 0,
        positionPoint: rectanglePrimitiveTestPoint2.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            0,
            0,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        angle: 0.7853982,
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    rectanglePrimitiveTestPoint2.model.primitives.push(rectanglePrimitiveTest2.model);
    rectanglePrimitiveTestPoint2.addChild(rectanglePrimitiveTest2);
    //</editor-fold>
    //<editor-fold desc="Test 3">
    // -- rectangle Primitive Test Point 3
    const rectanglePrimitiveTestPoint3 = new MedsurfDraw.PositionPoint(image, {
        id: 'rectanglePrimitiveTestPoint3',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 900 / image.dimensions.width,
            y: 300 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(rectanglePrimitiveTestPoint3);
    // -- rectangle Primitive Test 3
    const rectanglePrimitiveTest3 = new MedsurfDraw.RectanglePrimitive(image, {
        id: 'rectanglePrimitiveTest3',
        type: Models.ImageObjectType.RECTANGLE,
        isStatic: false,
        zIndex: 0,
        positionPoint: rectanglePrimitiveTestPoint3.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            10 / image.dimensions.width,
            10 / image.dimensions.width,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    rectanglePrimitiveTestPoint3.model.primitives.push(rectanglePrimitiveTest3.model);
    rectanglePrimitiveTestPoint3.addChild(rectanglePrimitiveTest3);
    //</editor-fold>
    //</editor-fold>
    //<editor-fold desc="Arrow Primitive">
    //<editor-fold desc="Test 1">
    // -- arrow Primitive Test Point 1
    const arrowPrimitiveTestPoint1 = new MedsurfDraw.PositionPoint(image, {
        id: 'arrowPrimitiveTestPoint1',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1000 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(arrowPrimitiveTestPoint1);
    // -- arrow Primitive Test 1
    const arrowPrimitiveTest1 = new MedsurfDraw.ArrowPrimitive(image, {
        id: 'arrowPrimitiveTest1',
        type: Models.ImageObjectType.ARROW,
        isStatic: false,
        zIndex: 0,
        positionPoint: arrowPrimitiveTestPoint1.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            0,
            0,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    arrowPrimitiveTestPoint1.model.primitives.push(arrowPrimitiveTest1.model);
    arrowPrimitiveTestPoint1.addChild(arrowPrimitiveTest1);
    //</editor-fold>
    //<editor-fold desc="Test 2">
    // -- arrow Primitive Test Point 2
    const arrowPrimitiveTestPoint2 = new MedsurfDraw.PositionPoint(image, {
        id: 'arrowPrimitiveTestPoint2',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1000 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(arrowPrimitiveTestPoint2);
    // -- arrow Primitive Test 2
    const arrowPrimitiveTest2 = new MedsurfDraw.ArrowPrimitive(image, {
        id: 'arrowPrimitiveTest2',
        type: Models.ImageObjectType.ARROW,
        isStatic: false,
        zIndex: 0,
        positionPoint: arrowPrimitiveTestPoint2.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            0,
            0,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        angle: 0.7853982,
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    arrowPrimitiveTestPoint2.model.primitives.push(arrowPrimitiveTest2.model);
    arrowPrimitiveTestPoint2.addChild(arrowPrimitiveTest2);
    //</editor-fold>
    //<editor-fold desc="Test 3">
    // -- arrow Primitive Test Point 3
    const arrowPrimitiveTestPoint3 = new MedsurfDraw.PositionPoint(image, {
        id: 'arrowPrimitiveTestPoint3',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1000 / image.dimensions.width,
            y: 300 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(arrowPrimitiveTestPoint3);
    // -- arrow Primitive Test 3
    const arrowPrimitiveTest3 = new MedsurfDraw.ArrowPrimitive(image, {
        id: 'arrowPrimitiveTest3',
        type: Models.ImageObjectType.ARROW,
        isStatic: false,
        zIndex: 0,
        positionPoint: arrowPrimitiveTestPoint3.model,
        rotateWithLine: false,
        rectangle: new PIXI.Rectangle(
            10 / image.dimensions.width,
            10 / image.dimensions.width,
            40 / image.dimensions.width,
            40 / image.dimensions.width
        ),
        strokeWidth: 1 / image.dimensions.width,
        options: {
            hasLine: true,
            lineColor: 0x000000,
            hasFill: true,
            fillColor: 0x00FF00,
            fillAlpha: 0.6
        }
    });
    // @ts-ignore
    arrowPrimitiveTestPoint3.model.primitives.push(arrowPrimitiveTest3.model);
    arrowPrimitiveTestPoint3.addChild(arrowPrimitiveTest3);
    //</editor-fold>
    //</editor-fold>
    //<editor-fold desc="Text Primitive">
    // -- text Primitive Test Point 1
    const textPrimitiveTestPoint1 = new MedsurfDraw.PositionPoint(image, {
        id: 'textPrimitiveTestPoint1',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1100 / image.dimensions.width,
            y: 100 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(textPrimitiveTestPoint1);
    // -- text Primitive Test Point 2
    const textPrimitiveTestPoint2 = new MedsurfDraw.PositionPoint(image, {
        id: 'textPrimitiveTestPoint2',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1100 / image.dimensions.width,
            y: 200 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(textPrimitiveTestPoint2);
    // -- text Primitive Test Point 3
    const textPrimitiveTestPoint3 = new MedsurfDraw.PositionPoint(image, {
        id: 'textPrimitiveTestPoint3',
        type: Models.ImageObjectType.POSITIONPOINT,
        isStatic: false,
        zIndex: 0,
        position: {
            x: 1100 / image.dimensions.width,
            y: 300 / image.dimensions.width,
        },
        isSelftest: true,
        selftestItems: [],
        primitives: []
    });
    image.addChild(textPrimitiveTestPoint3);
    //</editor-fold>
    //</editor-fold>

    //</editor-fold>

    image.sortChildren();
    app.stage.emit("setDefaultMode", "default");
    app.stage.interactive = true;

    return Promise.resolve();
}

/**
 * Start generator
 * @private
 */
function _startGenerator(): void {
    /* TODO not for copy */
    for(const layerImageObject of layerGroupsArray['sketch'].layerImageObjects || []) {
        const imageObject = image.getChildByName(layerImageObject.imageObject);
        if (imageObject) {
            // (imageObject as MedsurfDraw.DrawableObject).modeInteraction.setModeItem('draw_clickable');
            (imageObject as MedsurfDraw.DrawableObject).modeInteraction.setMode('draw_clickable');
        }
    }
    /**/
}

/**
 * Window ready
 */
ready(() => {
    WebFont.load({
        custom: {
            families: ["Font Awesome 5 Pro"],
            urls: ["./assets/fonts/fontawesome/css/all.css"],
        },
        // timeout: 2000,
        active: async () => {
            const width = canvasDiv.clientWidth;
            const height = canvasDiv.clientHeight;

            const currentCanvasAspect = { width: width / window.innerWidth, height: height / window.innerHeight };

            // Load cursors
            loader = PIXI.Loader.shared;
            // Setup loader
            loader.concurrency = 20;
            loader.add("transparent", "./assets/images/transparent.png");
            loader.load(async (loader: PIXI.Loader, resources: any) => {
                //<editor-fold desc="PIXI">
                PIXI.settings.PRECISION_FRAGMENT = PIXI.PRECISION.HIGH;
                // PIXI.settings.FAIL_IF_MAJOR_PERFORMANCE_CAVEAT = false;
                // PIXI.settings.ROUND_PIXELS = true;

                app = new PIXI.Application({
                    width: width,
                    height: height,
                    view: canvas,
                    antialias: true
                });
                //</editor-fold>

                //<editor-fold desc="PIXI plugins">
                /**
                 * setup gsap
                 */
                /**/
                PixiPlugin.registerPIXI(PIXI);
                gsap.registerPlugin(PixiPlugin);
                /**/

                /** TODO remove
                 * Setup viewport
                 * /
                const viewport = new Viewport({
                    screenWidth: window.innerWidth,
                    screenHeight: window.innerHeight,
                    worldWidth: width,
                    worldHeight: height,
                    interaction: app.renderer.plugins.interaction // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
                });

                // add the viewport to the stage
                app.stage.addChild(viewport);

                // activate plugins
                viewport
                    .drag()
                    .pinch()
                    .wheel();
                // TODO too fancy .decelerate();
                */

                /**
                 * setup cull
                 */
                /*
                const cull = new Simple() // new SpatialHash()
                cull.addList(image.children)
                cull.cull(image.getVisibleBounds())
                */
                /* TODO * /
                const cull = new PixiCull.Cull({ recursive: true, toggle: 'renderable' });
                cull.add(app.stage);
                /**/
                //</editor-fold>

                //<editor-fold desc="Window events">
                // Stop execution when not active
                window.addEventListener("visibilitychange", () => {
                    if (document.hidden) {
                        app.stop();
                    } else {
                        app.start();
                    }
                });
                // Resize and orientationchange events
                window.addEventListener("resize", (event: UIEvent) => {
                    app.renderer.resize(window.innerWidth * 0.7, window.innerHeight * 0.7);
                    app.stage.emit("resize");
                });
                window.addEventListener("orientationchange", (event: Event) => {
                    app.renderer.resize(window.innerWidth * 0.7, window.innerHeight * 0.7);
                    app.stage.emit("resize");
                });
                //</editor-fold>

                //<editor-fold desc="App Ticker">
                PIXI.Ticker.shared.add(() => { MedsurfDraw.Keyboard.update(); });
                // PIXI.Ticker.shared.add(() => { cull.cull(image.getVisibleBounds()) });
                // app.ticker.add(() => { MedsurfDraw.Keyboard.update(); });
                // app.ticker.add(() => { Keyboard.update(); });
                /*
                app.ticker.add((delta: number) => {
                    Keyboard.update();
                });
                */
                //</editor-fold>

                //<editor-fold desc="App Renderer">
                /** /
                app.renderer.on('prerender', () => {
                    cull.cull(app.renderer.screen);
                });
                /**/
                //</editor-fold>

                //<editor-fold desc="Button Events">
                buttonDefault.addEventListener("click", () => {
                    app.stage.emit("setDefaultMode", "default");
                });
                buttonMarker.addEventListener("click", () => {
                    app.stage.emit("setDefaultMode", "marker");
                });
                buttonSelftest.addEventListener("click", () => {
                    app.stage.emit("setDefaultMode", "selftest");
                });
                buttonAuthor.addEventListener("click", () => {
                    app.stage.emit("setDefaultMode", "author");
                });
                //</editor-fold>

                //<editor-fold desc="Cursor setup">
                const cursor = new PIXI.Text("\uf245", {
                    align: "left",
                    fill: 0x363636,
                    fontSize: 20,
                    fontFamily: "Font Awesome 5 Pro",
                    fontStyle: "normal",
                    fontVariant: "normal",
                    fontWeight: "900"
                });
                cursor.name = "cursor";
                cursor.visible = false;
                app.stage.addChild(cursor);

                console.log(app.renderer.plugins);
                const interaction = app.renderer.plugins.interaction;
                interaction.cursorStyles["default"] = (mode: any) => {
                    cursor.text = "\uf245";
                    cursor.anchor.set(0, 0);
                    cursor.rotation = 0;
                };
                interaction.cursorStyles["draw"] = (mode: any) => {
                    cursor.text = "\uf304";
                    cursor.anchor.set(0, 1);
                    cursor.rotation = 0;
                };
                interaction.cursorStyles["fill"] = (mode: any) => {
                    cursor.text = "\uf043";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = 0;
                };
                interaction.cursorStyles["move"] = (mode: any) => {
                    cursor.text = "\uf0b2";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = 0;
                };
                interaction.cursorStyles["rotate"] = (mode: any) => {
                    cursor.text = "\uf2ea";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = 0;
                };
                interaction.cursorStyles["scale-h"] = (mode: any) => {
                    cursor.text = "\uf337";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = ((interaction.eventData.currentTarget && interaction.eventData.currentTarget.parent && typeof interaction.eventData.currentTarget.parent.getRotation === "function") ? interaction.eventData.currentTarget.parent.getRotation() : 0);
                };
                interaction.cursorStyles["scale-v"] = (mode: any) => {
                    cursor.text = "\uf338";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = ((interaction.eventData.currentTarget && interaction.eventData.currentTarget.parent && typeof interaction.eventData.currentTarget.parent.getRotation === "function") ? interaction.eventData.currentTarget.parent.getRotation() : 0);
                };
                interaction.cursorStyles["scale-nesw"] = (mode: any) => {
                    cursor.text = "\uf424";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = ((interaction.eventData.currentTarget && interaction.eventData.currentTarget.parent && typeof interaction.eventData.currentTarget.parent.getRotation === "function") ? interaction.eventData.currentTarget.parent.getRotation() : 0);
                };
                interaction.cursorStyles["scale-nwse"] = (mode: any) => {
                    cursor.text = "\uf424";
                    cursor.anchor.set(0.5, 0.5);
                    cursor.rotation = Math.PI / 2 + ((interaction.eventData.currentTarget && interaction.eventData.currentTarget.parent && typeof interaction.eventData.currentTarget.parent.getRotation === "function") ? interaction.eventData.currentTarget.parent.getRotation() : 0);
                };
                interaction.cursorStyles["select"] = (mode: any) => {
                    cursor.text = "\uf25a";
                    cursor.anchor.set(0.3, 0);
                    cursor.rotation = 0;
                };
                interaction.on("pointerover", () => {
                    cursor.visible = true;
                });
                interaction.on("pointerout", () => {
                    cursor.visible = false;
                });
                interaction.on("pointermove", (event: PIXI.InteractionEvent) => {
                    cursor.position.set(event.data.global.x, event.data.global.y);
                    app.stage.setChildIndex(cursor, app.stage.children.length - 1);
                });
                //</editor-fold>

                //<editor-fold desc="Navigator">
                /* TODO check this with Deep Zoom
                let navigator = new MedsurfDraw.NavigatorElement({
                    renderer: app.renderer,
                    canvas: canvas,
                    image: image,
                    app: app
                });

                // Add navigator to stage
                app.stage.addChild(navigator);
                */
                //</editor-fold>

                //<editor-fold desc="VirtualPointer">
                const size = canvas.width > canvas.height ? canvas.width : canvas.height;
                virtualPointer = new MedsurfDraw.VirtualPointerElement({
                    rectangle: new PIXI.Rectangle(canvas.width / 2, canvas.height / 2, size / 10, size / 16)
                });
                virtualPointer.hideItem();

                // Add virtual pointer to stage
                app.stage.addChild(virtualPointer);
                //</editor-fold>

                await setupImage();
                await setupImageObjects();
            });
        },
    })
});
/**/

// Setup buttons
buttonVirtualPointer.addEventListener("click", () => {
    const size = canvas.width > canvas.height ? canvas.width : canvas.height;
    virtualPointer.rectangle = new PIXI.Rectangle(canvas.width / 2, canvas.height / 2, size / 10, size / 16);
    virtualPointer.draw();

    virtualPointer.toggleItem();
});
buttonTest.addEventListener("click", () => {
    // Add Middle point
    // line1-5; fillLine; realLine1
    (elementsArray['line1'] as MedsurfDraw.Line).onButtonAddMiddlePoint(image, layerGroupsArray['layerGroupModel1']);
    // (elementsArray['fillLine'] as MedsurfDraw.Line).onButtonAddMiddlePoint(image);
    // (elementsArray['realLine1'] as MedsurfDraw.Line).onButtonAddMiddlePoint(image, layerGroupsArray['realLayerGroup1']);

    // Check fill
    // (elementsArray['fill1'] as MedsurfDraw.FillCollection).onColor(MedsurfDraw.ColorMode.LINE, false, undefined, undefined);

    /* Text add middle point button * /
    line1.onButtonAddMiddlePoint(image);
    line2.onButtonAddMiddlePoint(image);
    line3.onButtonAddMiddlePoint(image);
    line4.onButtonAddMiddlePoint(image);
    line5.onButtonAddMiddlePoint(image);
    fillLine.onButtonAddMiddlePoint(image);
    /**/
    // line1.onColor(ColorMode.LINE, true, 0x000000, 0.3);

    // image.externControlSelectImageObject(arrowPoint.model);
    // image.externControlSelectPrimitive(arrowPoint.model, arrow.model);
    /*
    image.externControlSelectGrouping({
        id: 'grouping1',
        groupImages: [
            {
                id: 'groupImage1',
                image,
                layerGroupObjects: [
                    {
                        id: 'layerGroupObject1',
                        layerGroup: 'group1',
                        order: 0
                    },
                    {
                        id: 'layerGroupObject2',
                        layerGroup: 'group2',
                        order: 0
                    }
                ],
            }
        ],
        order: 0,
        name: 'test grouping',
    });
    */

    /*
    console.log("sketch", layerGroupsArray['sketch']);
    image.externControlSelectLayerGroup(layerGroupsArray['sketch']);
    */
    // elementsArray['text'].modeInteraction.setMode('delete');

    // Position point
    /** /
     const generator = new MedsurfDraw.PositionPointGenerator({
                        target: image,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        layerGroup: sketch,
                        image: {}
                    });
     /**/
    // Line
    /** /
     const generator = new MedsurfDraw.LineGenerator({
                        target: image,
                        // bezierCourve: false,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        // startElement: ellipsePoint,
                        layerGroup: sketch,
                        image: {}
                    });
     /**/
    // Polygon
    /** /
     const generator = new MedsurfDraw.PolygonGenerator({
                        target: image,
                        // bezierCourve: false,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        // startElement: test17Point,
                        layerGroup: layerGroupsArray['layerGroupModel1'],
                        image: {}
                    });
     /**/
    // Arrow primitive
    /** /
     const generator = new MedsurfDraw.ArrowPrimitiveGenerator({
                        target: image,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        // startElement: test17Point,
                        layerGroup: sketch,
                        image: {},
                        rotateWithLine: false,
                    });
     /**/
    // Ellipse primitive
    /** /
     const generator = new MedsurfDraw.EllipsePrimitiveGenerator({
                        target: image,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        // startElement: test17Point,
                        layerGroup: sketch,
                        image: {},
                        rotateWithLine: false,
                    });
     /**/
    // Rectangle primitive
    /** /
     const generator = new MedsurfDraw.RectanglePrimitiveGenerator({
                        target: image,
                        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
                        startElement: test17Point,
                        layerGroup: sketch,
                        image: {},
                        rotateWithLine: false,
                    });
     /**/
    // Text primitive
    /** /
    const generator = new MedsurfDraw.TextPrimitiveGenerator({
        target: image,
        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
        // startElement: test17Point,
        layerGroup: layerGroupsArray['sketch'],
        image: {},
        rotateWithLine: false,
    });
    /**/
    // Legend collection
    /** /
    const generator = new MedsurfDraw.LegendGenerator({
        target: image,
        layerGroup: layerGroupsArray['sketch'],
        image: {}
    });
    /**/
    // Fill collection
    /** /
    const generator = new MedsurfDraw.FillGenerator({
        target: image,
        bezierCourve: false,
        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
        // startElement: test17Point
        layerGroup: layerGroupsArray['sketch'],
        image: {},
    });
    /**/
    // Marker
    /** /
     const generator = new MedsurfDraw.MarkerGenerator({
            target: image,
            layerGroup: layerGroupsArray['sketch'],
            image: {}
        });
     /**/
    // Primitive Marker
    /** /
     const generator = new MedsurfDraw.PrimitiveMarkerGenerator({
                        target: image,
                        type: Models.ImageObjectType.RECTANGLE,
                        rotateWithLine: true,
                        layerGroup: sketch,
                        image: {}
                    });
     /**/
    // Arrow
    /** /
    const generator = new MedsurfDraw.ArrowGenerator({
        target: image,
        layerGroup: layerGroupsArray['sketch'],
        image: {}
    });
    /**/
    // Ellipse
    /** /
    const generator = new MedsurfDraw.EllipseGenerator({
        target: image,
        layerGroup: layerGroupsArray['sketch'],
        image: {}
    });
    /**/
    // Rectangle
    /** /
    const generator = new MedsurfDraw.RectangleGenerator({
        target: image,
        layerGroup: layerGroupsArray['sketch'],
        image: {}
    });
    /**/
    // Interactive area
    /** /
    const generator = new MedsurfDraw.InteractiveAreaGenerator({
        target: image,
        bezierCourve: false,
        stickMode: MedsurfDraw.StickMode.POSITIONPOINTS,
        // startElement: test17Point,
        layerGroup: layerGroupsArray['sketch'],
        image: {}
    });
    /**/
    // Select item
    /** /
     const generator = new MedsurfDraw.SelectItemGenerator({
            target: image,
            element: elementsArray['textPoint'] as MedsurfDraw.PositionPoint,
        });
     /**/
    // Clone item
    /** /
     const generator = new MedsurfDraw.CloneItemGenerator({
                        target: image,
                        imageObjects: image.getImageObjects().map((io) => (io as any).model),
                        element: (elementsArray['textPoint'] as MedsurfDraw.PositionPoint).model,
                        imageObjectMappings: [],
                        layerGroup: layerGroupsArray['sketch'],
                        image: {}
                    });
     // ellipsePoint, line1, fill1, legend
     /**/
    // Clone layer
    /** /
     const generator = new MedsurfDraw.CloneLayerGenerator({
                        target: image,
                        imageObjects: image.getImageObjects().map((io) => (io as any).model),
                        cloneLayerGroup: layerGroupsArray['layerGroupModel1'],
                        imageObjectMappings: [],
                        layerGroupMappings: [],
                        // TODO layerGroup: sketch,
                        image: {},
                    });
     /**/

    /** /
    // Default
    generator.once('startGenerator', _startGenerator, this); // NOT FOR CLONING; TODO layerGroup testing
    generator.once('endGenerator', generator.endGenerator, generator);
    generator.once('abortGenerator', generator.abortGenerator, generator);
    /**/

    // TODO remove
    /*
    generator.once('startGenerator', () => {
        test17Point.modeInteraction.setModeItem('draw_clickable');
    });
    */
    // TODO remove


    // TODO remove only test
    buttonModel.addEventListener("click", () => {
        // console.log(image.modeInteraction.lastMode);

        // @ts-ignore
        console.log("generator", generator);
    });
    /**/
});
buttonTest2.addEventListener("click", () => {
    console.log("test");
    // text.onColor(ColorMode.TEXT, true, 12852509, 1);
    // line1.onColor(ColorMode.LINE, true, 10032729, 1);
    /*
    console.log(image, image.modeInteraction.lastMode);
    console.log(ellipsePoint, ellipsePoint.modeInteraction.lastMode);
    console.log(ellipse, ellipse.modeInteraction.lastMode);
    console.log(backgroundRectangle, backgroundRectangle.modeInteraction.lastMode);
    */
    /*
    // @ts-ignore
    console.log(image._selectGroupElement);
    // @ts-ignore
    console.log(image._selectItemElement);
    console.log(image.getImageObjects());
    // @ts-ignore
    const objects = image.getImageObjects().filter((io) => io.data.model.type === Models.ImageObjectType.TEXT);
    for (let o of objects) {
        console.log(o);
    }
    */
    /** /
     image.externControlSelectPrimitive(ellipsePoint.model, ellipse.model);
    /**/
    /** /
    image.externControlSelectPrimitive((elementsArray['textPoint'] as MedsurfDraw.PositionPoint).model, (elementsArray['backgroundRectangle'] as MedsurfDraw.BackgroundRectanglePrimitive).model);
    /**/
    (elementsArray['backgroundRectangle'] as MedsurfDraw.BackgroundRectanglePrimitive).onButtonDelete(new PIXI.InteractionEvent());
});
buttonTest3.addEventListener("click", async () => {
    /*
    console.log(sketch);
    // @ts-ignore
    console.log(sketch.layerImageObjects.map(lio => lio.imageObject));
    // @ts-ignore
    console.log(image.getImageObjects().map(io => io.data.model));
    /**/
    /** /
     image.externControlSelectPrimitive(textPoint.model, text.model);
     /**/

    /**
     * Check image reset
     */
    /*
    console.log("start: change image");
    await destroyImage();
    await setupImage();
    await setupImageObjects();
    console.log("end: change image");
    */

    /**
     * Add Legend column
     */
    // (elementsArray['legend'] as MedsurfDraw.LegendCollection).onButtonAddLegendColumn();
    (elementsArray['text'] as MedsurfDraw.TextPrimitive).onButtonAddBackgroundRectangle();
});

/**
 * TODO TEST this
 * @param filterManager
 * @param input
 * @param output
 * @param clear
 */
/* TODO TEST
function blurFilterPassApply(filterManager: any, input: any, output: any, clear: any)
{
    if (output)
    {
        if (this.horizontal)
        {
            this.uniforms.strength = (1 / output.width) * (output.width / input.width);
        }
        else
        {
            this.uniforms.strength = (1 / output.height) * (output.height / input.height);
        }
    }
    else
    {
        if (this.horizontal) // eslint-disable-line
        {
            this.uniforms.strength = (1 / filterManager.renderer.width) * (filterManager.renderer.width / input.width);
        }
        else
        {
            this.uniforms.strength = (1 / filterManager.renderer.height) * (filterManager.renderer.height / input.height); // eslint-disable-line
        }
    }

    // screen space!
    this.uniforms.strength *= this.strength;
    this.uniforms.strength /= this.passes;

    if (this.passes === 1)
    {
        filterManager.applyFilter(this, input, output, clear);
    }
    else
    {
        const renderTarget = filterManager.getFilterTexture();
        const renderer = filterManager.renderer;

        let flip = input;
        let flop = renderTarget;

        this.state.blend = false;
        // changed false to true here
        filterManager.applyFilter(this, flip, flop, true);

        for (let i = 1; i < this.passes - 1; i++)
        {
            renderer.renderTexture.bind(flip, flip.filterFrame);

            this.uniforms.uSampler = flop;

            const temp = flop;

            flop = flip;
            flip = temp;

            renderer.shader.bind(this);
            renderer.geometry.draw(5);
        }

        this.state.blend = true;
        filterManager.applyFilter(this, flop, output, clear);
        filterManager.returnFilterTexture(renderTarget);
    }
}

PIXI.filters.BlurFilterPass.prototype.apply = blurFilterPassApply;
*/
