import * as MedsurfDraw from "../../public-api";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { v4 as uuidv4 } from 'uuid';
import {BaseContainer, BaseContainerImageModel} from "./BaseContainer";
import {Design} from "../../config/design";

/**
 * Number data model
 */
export class NumberDataModel {
    value: number | undefined;
    minValue: number;
    maxValue: number;
    gab: number;
}

export enum ColorMode {
    LINE,
    FILL,
    TEXT,
}

export enum NumberMode {
    WIDTH,
    SIZE,
    PAD,
}

/**
 * Base element container model
 */
export class BaseElementContainerModel<TModel extends Models.ImageObject> extends BaseContainerImageModel {
    model: TModel
}

/**
 * Base element container
 * Override default behaviour of base container
 */
export abstract class BaseElementContainer<TModel extends Models.ImageObject, TParams extends BaseElementContainerModel<TModel>> extends BaseContainer<TParams> {
    /**
     * Members
     */
    private _imageObjectMenu: MedsurfDraw.MenuEntryModel[];

    /**
     * Constructor
     * @param model
     */
    protected constructor(model: TParams) {
        // Parent constructor
        super(model);

        // Members
        this.imageObjectMenu = [];

        // Setup
        if (!this.model.id) {
            this.model.id = uuidv4();
        }
        this.name = this.model.id as string;
    }

    //<editor-fold desc="Menu">
    public getNumberData(): NumberDataModel {
        return {
            value: Design.default.numbers.defaultValue,
            minValue: Design.default.numbers.defaultMinValue,
            maxValue: Design.default.numbers.defaultMaxValue,
            gab: Design.default.numbers.defaultGab
        };
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get model
     */
    public get model(): TModel {
        return this.data.model;
    }

    /**
     * Set model
     * @param value
     */
    public set model(value: TModel) {
        this.data.model = value;
    }

    /**
     * Get menu
     */
    public get imageObjectMenu(): MedsurfDraw.MenuEntryModel[] {
        return this._imageObjectMenu;
    }

    /**
     * Set menu
     * @param value
     */
    public set imageObjectMenu(value: MedsurfDraw.MenuEntryModel[]) {
        this._imageObjectMenu = value;
    }
    //</editor-fold>
}