import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { BaseGraphics, BaseGraphicsModel } from "../../bases/elements/BaseGraphics";

/**
 * Arrow model
 */
export class ArrowModel extends BaseGraphicsModel {
    rectangle: PIXI.Rectangle;
    wedgePoint: Models.Point | undefined;
    options: Models.ElementOptions = {hasLine: true, lineColor: 0xFFFFFF, lineAlpha: 1, hasFill: false};
    lineWidth: number | undefined = 1;
}

/**
 * Arrow
 */
export class Arrow extends BaseGraphics<ArrowModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: ArrowModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        // this.cacheAsBitmap = false;

        this.clear();

        // Setup polygon
        const polygon: PIXI.Point[] = [];
        polygon.push(new PIXI.Point(this.rectangle.x, this.rectangle.y + this.rectangle.height / 2), new PIXI.Point(this.rectangle.x + this.rectangle.width, this.rectangle.y));
        if (this.wedgePoint) {
            polygon.push(new PIXI.Point(this.wedgePoint.x, this.wedgePoint.y));
        }
        polygon.push(new PIXI.Point(this.rectangle.x + this.rectangle.width, this.rectangle.y + this.rectangle.height), new PIXI.Point(this.rectangle.x, this.rectangle.y + this.rectangle.height / 2));

        // Draw
        if (this.options.hasLine) {
            this.lineStyle(this.lineWidth || 0, this.options.lineColor, this.options.lineAlpha, 0.5);
        }
        if (this.options.hasFill) {
            this.beginFill(this.options.fillColor, this.options.fillAlpha);
        }
        this.drawPolygon(polygon);
        this.endFill();

        // this.cacheAsBitmap = true;
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get wegde point
     */
    public get wedgePoint(): Models.Point | undefined {
        return this.data.wedgePoint;
    }

    /**
     * Set wegde point
     * @param value
     */
    public set wedgePoint(value: Models.Point | undefined) {
        this.data.wedgePoint = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number | undefined {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number | undefined) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}