import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { BaseGraphics, BaseGraphicsModel } from "../../bases/elements/BaseGraphics";

/**
 * Circle model
 */
export class CircleModel extends BaseGraphicsModel {
    circle: PIXI.Circle;
    options: Models.ElementOptions = {hasLine: true, lineColor: 0xFFFFFF, lineAlpha: 1, hasFill: false};
    lineWidth: number | undefined = 1;
}

/**
 * Circle
 */
export class Circle extends BaseGraphics<CircleModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: CircleModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        // this.cacheAsBitmap = false;

        this.clear();

        // Get scale
        // const scale = this.getImage().scale;

        // Draw circle
        if (this.options.hasLine) {
            // this.lineStyle((this.lineWidth || 1) * scale.x, this.options.lineColor, this.options.lineAlpha, 1);
            this.lineStyle(this.lineWidth || 1, this.options.lineColor, this.options.lineAlpha, 1);
        }
        if (this.options.hasFill) {
            this.beginFill(this.options.fillColor, this.options.fillAlpha);
        }
        // this.drawCircle(this.circle.x, this.circle.y, this.circle.radius * scale.x);
        this.drawCircle(this.circle.x, this.circle.y, this.circle.radius);
        this.endFill();

        // this.scale.set(1 / scale.x, 1 / scale.y);

        // this.cacheAsBitmapResolution = 1 / this.getImage().scale.x;
        // this.cacheAsBitmap = true;
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get circle
     */
    public get circle(): PIXI.Circle {
        return this.data.circle;
    }

    /**
     * Set circle
     * @param value
     */
    public set circle(value: PIXI.Circle) {
        this.data.circle = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number | undefined {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number | undefined) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}