import * as PIXI from "pixi.js-legacy";
import { BaseContainer, BaseContainerModel } from "../../bases/elements/BaseContainer";
import { Design } from "../../config/design";

export class KillboxModel extends BaseContainerModel {
    rectangle: PIXI.Rectangle;
}

/**
 * Selftest element
 */
export class Killbox extends BaseContainer<KillboxModel> {
    /**
     * Elements
     */
    protected _killboxElement: PIXI.Graphics;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: KillboxModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.selftestElement.zIndex;
    }

    /**
     * Init
     */
    public init(): void {
        // Setup
        this.interactive = true;

        // Position
        this.position.set(this.rectangle.x, this.rectangle.y);

        // Elements
        // -- Killbox element
        this._killboxElement = new PIXI.Graphics();
        this.addChild(this._killboxElement);
    }

    /**
     * Draw
     */
    public draw(): void {
        this.cacheAsBitmap = false;

        // Elements
        // -- Killbox element
        this._killboxElement.clear();
        this._killboxElement.lineStyle(Design.selftestElement.lineWidth, Design.selftestElement.lineColor, Design.selftestElement.lineAlpha);

        // Top left
        this._killboxElement.moveTo(this.rectangle.x, this.rectangle.y);
        this._killboxElement.lineTo(this.rectangle.x + Design.selftestElement.gab, this.rectangle.y);

        this._killboxElement.moveTo(this.rectangle.x, this.rectangle.y);
        this._killboxElement.lineTo(this.rectangle.x, this.rectangle.y + Design.selftestElement.gab);

        // Top right
        this._killboxElement.moveTo(this.rectangle.x + this.rectangle.width, this.rectangle.y);
        this._killboxElement.lineTo(this.rectangle.x + this.rectangle.width - Design.selftestElement.gab, this.rectangle.y);

        this._killboxElement.moveTo(this.rectangle.x + this.rectangle.width, this.rectangle.y);
        this._killboxElement.lineTo(this.rectangle.x + this.rectangle.width, this.rectangle.y + Design.selftestElement.gab);

        // Bottom left
        this._killboxElement.moveTo(this.rectangle.x, this.rectangle.y + this.rectangle.height);
        this._killboxElement.lineTo(this.rectangle.x + Design.selftestElement.gab, this.rectangle.y + this.rectangle.height);

        this._killboxElement.moveTo(this.rectangle.x, this.rectangle.y + this.rectangle.height);
        this._killboxElement.lineTo(this.rectangle.x, this.rectangle.y + this.rectangle.height - Design.selftestElement.gab);

        // Bottom right
        this._killboxElement.moveTo(this.rectangle.x + this.rectangle.width, this.rectangle.y + this.rectangle.height);
        this._killboxElement.lineTo(this.rectangle.x + this.rectangle.width - Design.selftestElement.gab, this.rectangle.y + this.rectangle.height);

        this._killboxElement.moveTo(this.rectangle.x + this.rectangle.width, this.rectangle.y + this.rectangle.height);
        this._killboxElement.lineTo(this.rectangle.x + this.rectangle.width, this.rectangle.y + this.rectangle.height - Design.selftestElement.gab);

        this._killboxElement.endFill();

        this.cacheAsBitmap = true;
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // Elements
        // -- Killbox element
        if (this._killboxElement) {
            this._killboxElement.destroy(options);
        }

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }
    //</editor-fold>
}