import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import { BaseGraphics, BaseGraphicsModel } from "../../bases/elements/BaseGraphics";

/**
 * Rectangle model
 */
export class RectangleModel extends BaseGraphicsModel {
    rectangle: PIXI.Rectangle;
    options: Models.ElementOptions = {hasLine: true, lineColor: 0xFFFFFF, lineAlpha: 1, hasFill: false};
    lineWidth: number | undefined = 1;
}

/**
 * Rectangle
 */
export class Rectangle extends BaseGraphics<RectangleModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: RectangleModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        // this.cacheAsBitmap = false;

        this.clear();

        // Draw rectangle
        if (this.options.hasLine) {
            this.lineStyle(this.lineWidth || 0, this.options.lineColor, this.options.lineAlpha, 1);
        }
        if (this.options.hasFill) {
            this.beginFill(this.options.fillColor, this.options.fillAlpha);
        }
        this.drawRect(this.rectangle.x, this.rectangle.y, this.rectangle.width, this.rectangle.height);
        this.endFill();

        // this.cacheAsBitmap = true;
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get rectangle
     */
    public get rectangle(): PIXI.Rectangle {
        return this.data.rectangle;
    }

    /**
     * Set rectangle
     * @param value
     */
    public set rectangle(value: PIXI.Rectangle) {
        this.data.rectangle = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number | undefined {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number | undefined) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}