import { BaseSprite, BaseSpriteModel } from "../../bases/elements/BaseSprite";
import { debounce } from 'debounce';

/**
 * Sprite
 */
export class Sprite extends BaseSprite<BaseSpriteModel> {
    /**
     * Methods
     */
    protected _debounceDrawMethod: EventListener;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: BaseSpriteModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Methods">
        this._debounceDrawMethod = debounce(this.draw.bind(this), 20).bind(this);
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Added
        this.on("added", () => {
            // Events
            // -- Debounce
            this.on("debounceDraw", this._debounceDrawMethod);
        });
        // -- Removed
        this.on("removed", () => {
            // Events
            // -- Debounce
            this.off("debounceDraw", this._debounceDrawMethod);
        });
        //</editor-fold>
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
    }
}