import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import { BaseText, BaseTextModel } from "../../bases/elements/BaseText";
import { Design } from "../../config/design";

/**
 * Text model
 */
export class TextModel extends BaseTextModel {
    text: string;
    style: PIXI.TextStyle;
}

/**
 * Text
 */
export class Text extends BaseText<TextModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: TextModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // On image zoom
        this.onImageZoom(1, 1);
    }

    //<editor-fold desc="ImageZoom">
    /**
     * On image zoom
     * @param scaleX
     * @param scaleY
     */
    public onImageZoom(scaleX: number, scaleY: number): void {
        // Update resolution
        let resolution = Math.round(scaleX);
        if (resolution < Design.text.minResolution) {
            resolution = Design.text.minResolution;
        }
        if (resolution > Design.text.maxResolution) {
            resolution = Design.text.maxResolution;
        }
        this.resolution = resolution;
    }
    //</editor-fold>

    //<editor-fold desc="BaseContainer">
    /**
     * Get rectangle
     */
    public getRectangle(): PIXI.Rectangle {
        try {
            if (this.parent instanceof MedsurfDraw.TextPrimitive || this.parent instanceof MedsurfDraw.LegendRow) {
                return this.parent.getRectangle();
            }
            return super.getRectangle();
        } catch (e) {
            // TODO what to do with this ???
            // console.warn(e);
            return new PIXI.Rectangle(this.position.x, this.position.y, 1, 1);
        }
    }

    /**
     * Get rotation
     */
    public getRotation(): number {
        if (this.parent instanceof MedsurfDraw.TextPrimitive || this.parent instanceof MedsurfDraw.LegendRow) {
            return this.parent.rotation;
        }
        return this.rotation;
    }
    //</editor-fold>
}