import * as PIXI from "pixi.js-legacy";
import * as Models from "@ascii-dev-user/medsurf-lib/models";
import * as MedsurfDraw from "../../public-api";
import {Design} from "../../config/design";
import {BaseContainer, BaseContainerImageModel} from "../../bases/elements/BaseContainer";

/**
 * Round button indicator element model
 */
export class RoundButtonIndicatorElementModel extends BaseContainerImageModel {
    circle?: PIXI.Circle;
    options?: Models.ElementOptions;
    lineWidth?: number;
}

/**
 * Round button indicator element
 */
export class RoundButtonIndicatorElement extends BaseContainer<RoundButtonIndicatorElementModel> {
    /**
     * Elements
     */
    protected _arcElement: MedsurfDraw.Arc;

    /**
     * Constructor
     * @param model
     */
    public constructor(model: RoundButtonIndicatorElementModel) {
        // Parent constructor
        super(model);

        // Setup
        this.zIndex = Design.menuButton.zIndex + 1;

        // Position
        this.position.set(this.circle.x, this.circle.y);
    }

    /**
     * Init
     */
    public init(parent: MedsurfDraw.ImageObject): void {
        // Setup
        this.interactive = true;

        /*
        let scale = this.imageScale.x * Design.roundButtonIndicatorElement.scaleAspect;
        let buttonRadius = this.circle.radius;
        buttonRadius = Math.floor(buttonRadius / scale);
        */
        let lineWidth = this.lineWidth;
        if (lineWidth < Design.roundButtonIndicatorElement.minLineWidth) {
            lineWidth = Design.roundButtonIndicatorElement.minLineWidth;
        }
        //<editor-fold desc="Elements">
        // -- Arc element
        this._arcElement = new MedsurfDraw.Arc({
            center: new PIXI.Point(0, 0),
            options: this.options,
            radius: this.circle.radius,
            lineWidth: lineWidth,
        });
        this._arcElement.angle = 225;
        this.addChild(this._arcElement);
        //</editor-fold>

        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        /*
        let scale = this.imageScale.x * Design.roundButtonIndicatorElement.scaleAspect;
        let buttonRadius = this.circle.radius;
        buttonRadius = Math.floor(buttonRadius / scale);
        */
        let lineWidth = this.lineWidth;
        if (lineWidth < Design.roundButtonIndicatorElement.minLineWidth) {
            lineWidth = Design.roundButtonIndicatorElement.minLineWidth;
        }
        //<editor-fold desc="Elements">
        // -- Arc element
        this._arcElement.radius = this.circle.radius;
        this._arcElement.lineWidth = lineWidth;
        this._arcElement.draw();
        //</editor-fold>
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        //<editor-fold desc="Elements">
        // -- Arc element
        if (this._arcElement) {
            this._arcElement.destroy(options);
        }
        //</editor-fold>

        // Parent destroy
        super.destroy(options);
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get circle
     */
    public get circle(): PIXI.Circle {
        return this.data.circle || new PIXI.Circle(0, 0, 20);
    }

    /**
     * Set circle
     * @param value
     */
    public set circle(value: PIXI.Circle) {
        this.data.circle = value;

        // Elements
        // -- Arc element
        this._arcElement.radius = this.circle.radius;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options || {hasLine: true, lineColor: Design.roundButtonIndicatorElement.lineColor, lineAlpha: Design.roundButtonIndicatorElement.lineAlpha, hasFill: false};
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;

        // Elements
        // -- Arc element
        this._arcElement.options = this.options;

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Get line width
     */
    public get lineWidth(): number {
        return this.data.lineWidth || Design.roundButtonIndicatorElement.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number) {
        this.data.lineWidth = value;

        // Elements
        // -- Arc element
        this._arcElement.lineWidth = this.lineWidth;

        // Draw
        this.emit("debounceDraw");
    }
    //</editor-fold>
}