import * as Models from "@ascii-dev-user/medsurf-lib/models";
import * as MedsurfDraw from "../../public-api";
import {BaseGraphics, BaseGraphicsImageModel} from "../../bases/elements/BaseGraphics";

/**
 * Line corner model
 */
export class LineCornerModel extends BaseGraphicsImageModel {
    points: MedsurfDraw.StraightLinePointsModel | MedsurfDraw.QuadraticCurvePointsModel | MedsurfDraw.BezierLinePointsModel | undefined;
    options: Models.ElementOptions;
    lineWidth: number
}

/**
 * Line corner
 */
export class LineCorner extends BaseGraphics<LineCornerModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: LineCornerModel) {
        // Parent constructor
        super(model);
    }

    /**
     * Init
     */
    public init(): void {
        // Sort items
        this.sortChildren();

        // Draw
        this.emit("debounceDraw");
    }

    /**
     * Draw
     */
    public draw(): void {
        // @ts-ignore
        if (!this._geometry) {
            return;
        }

        this.cacheAsBitmap = false;

        this.clear();

        // Draw corner
        if (this.points) {
            this.lineStyle(this.lineWidth / 2, this.options.lineColor, this.options.lineAlpha);
            this.beginFill(this.options.lineColor, this.options.lineAlpha);
            this.drawCircle(this.points.start.x, this.points.start.y, this.lineWidth / 4);
            this.drawCircle(this.points.end.x, this.points.end.y, this.lineWidth / 4);

            this.endFill();
        }

        this.cacheAsBitmap = true;
    }

    //<editor-fold desc="Getter and Setter">
    /**
     * Get points
     */
    public get points(): MedsurfDraw.StraightLinePointsModel | MedsurfDraw.QuadraticCurvePointsModel | MedsurfDraw.BezierLinePointsModel | undefined {
        return this.data.points;
    }

    /**
     * Set points
     * @param value
     */
    public set points(value: MedsurfDraw.StraightLinePointsModel | MedsurfDraw.QuadraticCurvePointsModel | MedsurfDraw.BezierLinePointsModel | undefined) {
        this.data.points = value;
    }

    /**
     * Get options
     */
    public get options(): Models.ElementOptions {
        return this.data.options;
    }

    /**
     * Set options
     * @param value
     */
    public set options(value: Models.ElementOptions) {
        this.data.options = value;
    }

    /**
     * Get line width
     */
    public get lineWidth(): number {
        return this.data.lineWidth;
    }

    /**
     * Set line width
     * @param value
     */
    public set lineWidth(value: number) {
        this.data.lineWidth = value;
    }
    //</editor-fold>
}