import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../../public-api";
import {BaseGenerator, BaseGeneratorModel} from "../../bases/generators/BaseGenerator";

/**
 * Select item generator model
 */
export class SelectItemGeneratorModel extends BaseGeneratorModel<MedsurfDraw.Image> {
    element: MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection;
}

/**
 * Select item generator
 */
export class SelectItemGenerator extends BaseGenerator<MedsurfDraw.Image, SelectItemGeneratorModel> {
    /**
     * Constructor
     * @param model
     */
    public constructor(model: SelectItemGeneratorModel) {
        // Parent constructor
        super(model);

        //<editor-fold desc="Elements">
        // -- Image
        this.target.modeInteraction.setModeItem('choosing');
        // -- Element
        this.element.controlChooseItem();
        //</editor-fold>

        //<editor-fold desc="Events">
        // -- Image
        this.target.parent.on("rightup", this.chooseInteraction.abortChoose, this.chooseInteraction);
        this.target.parent.on("mouseup", this.chooseInteraction.endChoose, this.chooseInteraction);
        // -- Choose
        this.chooseInteraction.on("abortChoose", this.abortChoose, this);
        //</editor-fold>
    }

    /**
     * Start
     * @param event
     */
    public start(event: PIXI.InteractionEvent): void {
        // Parent
        super.start(event);
    }

    /**
     * End
     */
    public end(): void {
        //<editor-fold desc="Events">
        if (this.chooseInteraction) {
            // -- Image
            this.target.parent.off("rightup", this.chooseInteraction.abortChoose, this.chooseInteraction);
            this.target.parent.off("mouseup", this.chooseInteraction.endChoose, this.chooseInteraction);
            // -- Choose
            this.chooseInteraction.removeAllListeners();
        }
        //</editor-fold>

        // Parent end
        super.end();
    }

    /**
     * Destroy
     * @param options
     */
    public destroy(options?: { children?: boolean; texture?: boolean; baseTexture?: boolean }): void {
        // End
        this.end();
    }

    //<editor-fold desc="Extern">
    /**
     * End generator
     */
    public endGenerator(): void {
        // Parent end generator
        super.endGenerator();

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }

    /**
     * Abort generator
     */
    public abortGenerator(): void {
        // Parent abort generator
        super.abortGenerator();

        // Remove new elements
        this.target.getImageObjects()
            .filter((imageObject) => imageObject.modeInteraction.lastMode.startsWith('choose_new'))
            .forEach((imageObject) => {
                // Remove element
                imageObject.modeInteraction.setMode("delete_draw");
            });

        // Sort children
        this.target.sortChildren();

        // Set image default mode
        this.target.modeInteraction.setMode(this.target.modeInteraction.defaultMode);
    }
    //</editor-fold>

    //<editor-fold desc="Choose">
    /**
     * Abort generator
     * @param event
     */
    public abortChoose(event: PIXI.InteractionEvent): void {
        event.stopPropagation();

        // End
        this.end();

        // Emit abort generator
        this.emit("abortGenerator");
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get choose interaction
     */
    public get chooseInteraction(): MedsurfDraw.ChooseInteraction<MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection> {
        return this.element.chooseInteraction;
    }

    /**
     * Get element
     */
    public get element(): MedsurfDraw.PositionPoint | MedsurfDraw.FillCollection {
        return this.data.element;
    }
    //</editor-fold>
}