const Events = require('nom-events');

class CKeyboard {
    public keyStates: Map<any, any>;
    public events: any;

    constructor() {
        this.keyStates = new Map();
        this.events = new Events();
    }

    clear() {
        this.keyStates.clear();
    }

    update() {
        this.keyStates.forEach((value, keyCode) => {
            const event = this.keyStates.get(keyCode);

            event.alreadyPressed = true;

            if (event.wasReleased) {
                this.keyStates.delete(keyCode);
            }

            Keyboard.events.call('down', keyCode, event);
            Keyboard.events.call('down_' + keyCode, keyCode, event);
        });
    }

    isKeyDown(...args: any[]) {
        let result = false;
        for(let keyCode of args) {
            const event = this.keyStates.get(keyCode);
            if (event && !event.wasReleased)
                result = true;
        }

        return result;
    }

    isKeyUp(...args: any[]) {
        return !this.isKeyDown(args);
    }

    isKeyPressed(...args: any[]) {
        let result = false;

        if (args.length == 0)
            return false;

        for(let keyCode of args) {
            const event = this.keyStates.get(keyCode);
            if (event && !event.wasReleased && !event.alreadyPressed)
                result = true;
        }

        return result;
    }

    isKeyReleased(...args: any[]) {
        let result = false;

        if (args.length == 0)
            return false;

        for(let keyCode of args) {
            const event = this.keyStates.get(keyCode);
            if (event && event.wasReleased)
                result = true;
        }

        return result;
    }
}

const Keyboard = new CKeyboard();

window.addEventListener(
    "keydown", (event) => {
        if (!Keyboard.keyStates.get(event.code)) {
            Keyboard.keyStates.set(event.code, event);
            Keyboard.events.call('pressed', event.code, event);
            Keyboard.events.call('pressed_' + event.code, event.code, event);
        }
    }, false
);

window.addEventListener(
    "keyup", (event) => {
        event = Keyboard.keyStates.get(event.code);
        if (event) {
            //keyboard.keyStates.set(event.code, event);
            // @ts-ignore
            event.wasReleased = true;
            Keyboard.events.call('released', event.code, event);
            Keyboard.events.call('released_' + event.code, event.code, event);
        }
    }, false
);

export { Keyboard };

