import * as PIXI from "pixi.js-legacy";
import {BaseInteraction} from "../bases/interactions/BaseInteraction";

/**
 * Clone interaction
 */
export class CloneInteraction extends BaseInteraction {
    /**
     * Members
     */
    protected _mouseDown: number;
    protected _moveLock: boolean;

    /**
     * Constructor
     */
    public constructor() {
        // Parent constructor
        super();

        // Members
        this._mouseDown = 0;
        this._moveLock = false;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._mouseDown = 0;
        this._moveLock = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Start draw
     * @param event
     */
    public startClone(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this._mouseDown++;
        this.emit("startClone", event);
    }

    /**
     * Reset clone
     * @param event
     */
    public resetClone(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this.reset();
        this.emit("resetClone", event);
    }

    /**
     * On clone
     * @param event
     */
    public onClone(event: PIXI.InteractionEvent) {
        if (this._mouseDown) {
            this._moveLock = true;
            return;
        }
        this.emit("onClone", event);
    }

    /**
     * End clone
     * @param event
     */
    public endClone(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this._mouseDown--;
        if (this._moveLock) {
            this.resetClone(event);
            return;
        }
        this.emit("endClone", event);
    }

    /**
     * Abort clone
     * @param event
     */
    public abortClone(event: PIXI.InteractionEvent) {
        event.stopPropagation();
        this.reset();
        this.emit("abortClone", event);
    }
    //</editor-fold>
}