import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Mode interaction
 */
export class ModeInteraction<TElement extends MedsurfDraw.ImageObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    private _defaultMode: string;
    protected _parentImageObject: MedsurfDraw.ImageObject;
    protected _lastMode: string;

    /**
     * Constructor
     * @param element
     * @param defaultMode
     */
    public constructor(element: TElement, defaultMode: string = "init") {
        // Parent constructor
        super(element);

        // Members
        this.defaultMode = defaultMode;

        // Elements
        // -- Target element
        this.element.on("added", (parentImageObject: MedsurfDraw.ImageObject) => {
            this._parentImageObject = parentImageObject;
            this._lastMode = this.defaultMode;
            this.emit(this.defaultMode, this._parentImageObject, this.lastMode);
        });
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        return;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * Set default mode and change to it
     * @param mode
     * @param args
     */
    public setDefaultMode(mode: string, ...args: any[]): void {
        // Set default mode
        this.defaultMode = mode;

        // Change mode
        this._changeMode(mode);

        // Emit set default mode
        this.emit("setDefaultMode", mode);
    }

    /**
     * Set mode
     * @param mode
     * @param args
     */
    public setMode(mode: string, ...args: any[]): void {
        // Change mode
        this._changeMode(mode);

        // Emit set mode
        this.emit("setMode", mode);
    }

    /**
     * Set mode children
     * @param mode
     */
    public setModeChildren(mode: string): void {
        // Emit set mode
        this.emit("setMode", mode);
    }

    /**
     * Set mode item
     * @param mode
     * @param args
     */
    public setModeItem(mode: string, ...args: any[]): void {
        // Change mode
        this._changeMode(mode);
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Change mode
     * @param mode
     * @param args
     * @private
     */
    protected _changeMode(mode: string, ...args: any[]) {
        // TODO remove
        /*
        if (this.element.name && this.element.name.indexOf('ine') !== -1) {
            console.log(this.element.name, this.lastMode, mode);
        }
        */
        this.emit("remove-" + this.lastMode, this._parentImageObject, mode);
        this._lastMode = mode;
        this.emit(mode, this._parentImageObject, mode, ...args);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    /**
     * Get default mode
     */
    public get defaultMode(): string {
        return this._defaultMode;
    }

    /**
     * Set default mode
     * @param value
     */
    public set defaultMode(value: string) {
        this._defaultMode = value;
    }

    /**
     * Get last mode
     */
    public get lastMode(): string {
        return this._lastMode;
    }
    //</editor-fold>
}
