import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

export const SELECT_HITTEST_EXTENSION: number = 1000;

/**
 * Select interaction
 */
export class SelectInteraction<TElement extends MedsurfDraw.ImageObject> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    protected _isSelected: boolean;

    /**
     * Constructor
     * @param element
     */
    public constructor(element: TElement) {
        // Parent constructor
        super(element);
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._isSelected = false;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * On pointer down
     * @param event
     */
    public onPointerDown(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
    }

    /**
     * On hover
     * @param event
     */
    public onHover(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "select";
        this.emit("onHover", event);
    }

    /**
     * On out
     * @param event
     */
    public onOut(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "default";
        this.emit("onOut", event);
    }

    /**
     * On select
     * @param event
     */
    public onSelect(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "select";
        this.emit("onSelectStart", event);

        if (this._isSelected) {
            return;
        }

        this._isSelected = true;

        this.emit("onSelect", event);
    }

    /**
     * On selftest
     * @param event
     */
    public onSelftest(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "select";
        this.emit("onSelftestStart", event);

        if (this._isSelected) {
            return;
        }

        this._isSelected = true;

        this.emit("onSelftest", event);
    }

    /**
     * On interactive
     * @param event
     */
    public onInteractive(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "select";
        this.emit("onInteractiveStart", event);

        if (this._isSelected) {
            return;
        }

        this._isSelected = true;

        this.emit("onInteractive", event);
    }

    /**
     * On release
     * @param event
     */
    public onRelease(event: PIXI.InteractionEvent): void {
        event.stopPropagation();
        this.element.cursor = "default";
        this.emit("onReleaseStart", event);

        if (!this._isSelected) {
            return;
        }

        this._isSelected = false;

        this.emit("onRelease", event);
    }
    //</editor-fold>
}