import * as MedsurfDraw from "../public-api";
import { BaseElementInteraction } from "../bases/interactions/BaseElementInteraction";

/**
 * Write interaction
 * from: https://github.com/limikael/PixiTextInput/blob/master/src/PixiTextInput.js#L417
 * with: https://www.npmjs.com/package/pixi.js-keyboard
 */
export class WriteInteraction<TElement extends MedsurfDraw.EditableText> extends BaseElementInteraction<TElement> {
    /**
     * Constructor
     * @param element
     */
    public constructor(element: TElement) {
        // Parent constructor
        super(element);
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        return;
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * On key down
     * @param keyCode
     * @param event
     */
    public onKeyDown(keyCode: string, event: KeyboardEvent): void {
    }

    /**
     * On key pressed
     * @param keyCode
     * @param event
     */
    public onKeyPressed(keyCode: string, event: KeyboardEvent): void {
        if (event.ctrlKey && (event.key === "c" || event.key === "3" || event.key === "v" || event.key === "F7" || event.key === "x" || event.key === "F9")) {
            return;
        }
        event.preventDefault();
        switch (event.key) {
            // 8
            case "Backspace":
                if (this.element.textSelection) {
                    this.deleteSelectedText();
                } else if (this.element.startTextIndex > 0) {
                    this.emit("onText", this.element.text.substring(0, this.element.startTextIndex - 1) + this.element.text.substring(this.element.startTextIndex));
                    this.moveCaretLeft(false);

                    /* TODO remove
                    // Update
                    this.element.getImage().updateImageObject(this.element.model);
                    */
                }
                return;
            // 9
            case "Tab":
                return;
            // 13
            case "Enter":
                if (event.shiftKey) {
                    if (this.element.textSelection) {
                        this.replaceSelectedText("\n");
                    } else {
                        this.emit("onText", this.element.text.substring(0, this.element.startTextIndex) + "\n" + this.element.text.substring(this.element.startTextIndex));
                        this.moveCaretRight(false);

                        /* TODO remove
                        // Update
                        this.element.getImage().updateImageObject(this.element.model);
                        */
                    }
                } else {
                    this.emit("endText");
                }
                return;
            // 16
            case "Shift":
                return;
            // 17
            case "Control":
                return;
            // 18
            case "Alt":
                return;
            // 19
            case "Pause":
                return;
            // 20
            case "CapsLock":
                return;
            // 27 TODO
            case "Escape":
                this.emit("endText");
                // TODO stop writing
                return;
            // 33
            case "PageUp":
                return;
            // 34
            case "PageDown":
                return;
            // 35
            case "End":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                if (event.ctrlKey && this.element.startTextIndex + 1 < this.element.text.length) {
                    this.element.startTextIndex = this.element.text.length;
                } else {
                    let nextPosition = this.element.text.indexOf("\n", this.element.startTextIndex);
                    if (nextPosition !== this.element.startTextIndex) {
                        this.element.startTextIndex = (nextPosition != -1) ? nextPosition : this.element.text.length;
                    } else {
                        this.moveCaretRight(event.shiftKey);
                    }
                }

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");
                return;
            // 36
            case "Home":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                if (event.ctrlKey && this.element.startTextIndex - 1 > 0) {
                    this.element.startTextIndex = 0;
                } else {
                    let nextPosition = this.element.text.lastIndexOf("\n", this.element.startTextIndex - 1);
                    if (nextPosition + 1 !== this.element.startTextIndex) {
                        this.element.startTextIndex = (nextPosition != -1) ? nextPosition + 1 : 0;
                    } else {
                        this.moveCaretLeft(event.shiftKey);
                    }
                }

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");
                return;
            // 37
            case "ArrowLeft":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                if (event.ctrlKey && this.element.startTextIndex - 1 > 0) {
                    let nextPosition = this.element.text.lastIndexOf("\n", this.element.startTextIndex - 1);
                    if (nextPosition + 1 !== this.element.startTextIndex) {
                        this.element.startTextIndex = (nextPosition != -1) ? nextPosition + 1 : 0;
                    } else {
                        this.moveCaretLeft(event.shiftKey);
                    }
                } else {
                    this.moveCaretLeft(event.shiftKey);
                }

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");

                return;
            // 38
            case "ArrowUp":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                this.moveCaretUp(event.shiftKey);

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");
                return;
            // 39
            case "ArrowRight":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                if (event.ctrlKey && this.element.startTextIndex + 1 < this.element.text.length) {
                    let nextPosition = this.element.text.indexOf("\n", this.element.startTextIndex);
                    if (nextPosition !== this.element.startTextIndex) {
                        this.element.startTextIndex = (nextPosition != -1) ? nextPosition : this.element.text.length;
                    } else {
                        this.moveCaretRight(event.shiftKey);
                    }
                } else {
                    this.moveCaretRight(event.shiftKey);
                }

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");

                return;
            // 40
            case "ArrowDown":
                if (event.shiftKey && !this.element.textSelection) {
                    this.element.textSelection = true;
                    this.element.endTextIndex = this.element.startTextIndex;
                }

                this.moveCaretDown(event.shiftKey);

                if (!event.shiftKey) {
                    this.element.textSelection = false;
                }

                this.emit("onSelection");
                this.emit("onCaret");
                return;
            // 45 TODO
            case "Insert":
                // TODO override
                return;
            // 46
            case "Delete":
                if (this.element.textSelection) {
                    this.deleteSelectedText();
                } else {
                    this.emit("onText", this.element.text.substring(0, this.element.startTextIndex) + this.element.text.substring(this.element.startTextIndex + 1));
                    this.emit("onCaret");

                    /* TODO remove
                    // Update
                    this.element.getImage().updateImageObject(this.element.model);
                    */
                }
                return;
            // 65
            case "a":
                if (event.ctrlKey) {
                    this.selectAllText();
                    return;
                }
                break;
            // 91
            case "OS":
                return;
            // 93
            case "ContextMenu":
                return;
            // 112 - 123
            case "F1":
            case "F2":
            case "F3":
            case "F4":
            case "F5":
            case "F6":
            case "F7":
            case "F8":
            case "F9":
            case "F10":
            case "F11":
            case "F12":
                return;
            // 144
            case "NumLock":
                return;
            // 145
            case "ScrollLock":
                return;
            // 225
            case "AltGraph":
                return;
        }

        if (this.element.textSelection) {
            this.replaceSelectedText(event.key);
        } else {
            this.emit("onText", this.element.text.substring(0, this.element.startTextIndex) + event.key + this.element.text.substring(this.element.startTextIndex));
            this.moveCaretRight(false);

            /* TODO remove
            // Update
            this.element.getImage().updateImageObject(this.element.model);
            */
        }
    }

    // TODO
    public onKeyReleased(keyCode: string, event: KeyboardEvent): void {
    }

    /**
     * On copy
     * @param event
     */
    public onCopy(event: ClipboardEvent): void {
        event.preventDefault();

        // @ts-ignore
        event.clipboardData.setData("text/plain", this.getSelectedText());
    }

    /**
     * On cut
     * @param event
     */
    public onCut(event: ClipboardEvent): void {
        event.preventDefault();

        // @ts-ignore
        event.clipboardData.setData("text/plain", this.getSelectedText());

        if (this.element.textSelection) {
            this.deleteSelectedText();
            this.moveCaretRight(false);
        }
    }

    /**
     * On paste
     * @param event
     */
    public onPaste(event: ClipboardEvent): void {
        event.preventDefault();

        // @ts-ignore
        const text = event.clipboardData.getData("text/plain");

        if (this.element.textSelection) {
            this.deleteSelectedText();
            this.moveCaretRight(false);
        }

        this.replaceSelectedText(text);
    }
    //</editor-fold>

    //<editor-fold desc="Public functions">
    /**
     * Select end of text
     */
    public selectEndText(): void {
        this.element.startTextIndex = this.element.text.length;
        this.emit("onCaret");
    }

    /**
     * Select all text
     */
    public selectAllText(): void {
        this.element.textSelection = true;
        this.element.startTextIndex = this.element.text.length;
        this.element.endTextIndex = 0;
        this.emit("onSelection");
    }
    //</editor-fold>

    //<editor-fold desc="Private functions">
    /**
     * Delete selected text
     */
    protected deleteSelectedText(): void {
        let startPosition: number;
        let endPosition: number;

        if (this.element.startTextIndex > this.element.endTextIndex) {
            startPosition = this.element.endTextIndex;
            endPosition = this.element.startTextIndex;
        } else {
            startPosition = this.element.startTextIndex;
            endPosition = this.element.endTextIndex;
        }

        if (startPosition === 0 && endPosition === this.element.text.length) {
            this.emit("onText", "");
        } else {
            this.emit("onText", this.element.text.substring(0, startPosition) + this.element.text.substring(endPosition));
        }

        this.element.textSelection = false;
        this.element.startTextIndex = startPosition;
        if (this.element.startTextIndex < 0) {
            this.element.startTextIndex = 0;
        }

        this.emit("onSelection");
        this.emit("onCaret");

        /* TODO remove
        // Update
        this.element.getImage().updateImageObject(this.element.model);
        */
    }

    /**
     * Replace selected text
     * @param text
     */
    protected replaceSelectedText(text: string): void {
        let startPosition: number;
        let endPosition: number;

        if (this.element.startTextIndex > this.element.endTextIndex) {
            startPosition = this.element.endTextIndex;
            endPosition = this.element.startTextIndex;
        } else {
            startPosition = this.element.startTextIndex;
            endPosition = this.element.endTextIndex;
        }

        if (startPosition === 0 && endPosition === this.element.text.length) {
            this.emit("onText", text);
        } else {
            this.emit("onText", this.element.text.substring(0, startPosition) + text + this.element.text.substring(endPosition));
        }

        this.element.textSelection = false;
        this.element.startTextIndex = startPosition + 1;
        if (this.element.startTextIndex < 0) {
            this.element.startTextIndex = 0;
        }

        this.emit("onSelection");
        this.emit("onCaret");

        /* TODO remove
        // Update
        this.element.getImage().updateImageObject(this.element.model);
        */
    }

    /**
     * Move caret left
     * @param shiftDown
     */
    protected moveCaretLeft(shiftDown: boolean): void {
        if (this.element.textSelection && !shiftDown) {
            if (this.element.startTextIndex > this.element.endTextIndex) {
                this.element.startTextIndex = this.element.endTextIndex;
            }
        } else {
            this.element.startTextIndex--;
            if (this.element.startTextIndex < 0) {
                this.element.startTextIndex = 0;
            }
        }

        this.emit("onCaret");
    }

    /**
     * Move caret up
     * @param shiftDown
     */
    protected moveCaretUp(shiftDown: boolean): void {
        let linePosition = this.element.text.lastIndexOf("\n", this.element.startTextIndex - 1);
        let previousLinePosition = this.element.text.lastIndexOf("\n", linePosition - 1);
        let leftIndex = Math.abs(this.element.startTextIndex - linePosition);

        if (this.element.startTextIndex <= leftIndex) {
            this.element.startTextIndex = 0;
        } else if ((this.element.startTextIndex - linePosition) > linePosition - previousLinePosition) {
            this.element.startTextIndex = linePosition;
        } else {
            this.element.startTextIndex = previousLinePosition + leftIndex;
        }

        this.emit("onCaret");
    }

    /**
     * Move caret right
     * @param shiftDown
     */
    protected moveCaretRight(shiftDown: boolean): void {
        if (this.element.textSelection && !shiftDown) {
            if (this.element.startTextIndex < this.element.endTextIndex) {
                this.element.startTextIndex = this.element.endTextIndex;
            }
        } else {
            this.element.startTextIndex++;
            if (this.element.startTextIndex > this.element.text.length) {
                this.element.startTextIndex = this.element.text.length;
            }
        }

        this.emit("onCaret");
    }

    /**
     * Move caret down
     * @param shiftDown
     */
    protected moveCaretDown(shiftDown: boolean): void {
        let linePosition = this.element.text.lastIndexOf("\n", this.element.startTextIndex - 1);
        let nextLinePosition = this.element.text.indexOf("\n", this.element.startTextIndex);
        let leftIndex = Math.abs(this.element.startTextIndex - linePosition);

        if (nextLinePosition === -1) {
            this.element.startTextIndex = this.element.text.length;
        } else if (this.element.startTextIndex == nextLinePosition) {
            this.element.startTextIndex = nextLinePosition + (nextLinePosition - linePosition);
        } else {
            this.element.startTextIndex = nextLinePosition + leftIndex;
        }

        this.emit("onCaret");
    }

    /**
     * Get selected text
     */
    protected getSelectedText(): string {
        if (this.element.textSelection) {
            if (this.element.startTextIndex < this.element.endTextIndex) {
                return this.element.text.substring(this.element.startTextIndex, this.element.endTextIndex);
            } else if (this.element.startTextIndex > this.element.endTextIndex) {
                return this.element.text.substring(this.element.endTextIndex, this.element.startTextIndex);
            }
        }
        return "";
    }
    //</editor-fold>
}