import * as PIXI from "pixi.js-legacy";
import * as MedsurfDraw from "../public-api";
import {BaseElementInteraction} from "../bases/interactions/BaseElementInteraction";

/**
 * Zoom interaction
 */
export class ZoomInteraction<TElement extends MedsurfDraw.Image | MedsurfDraw.NavigatorElement> extends BaseElementInteraction<TElement> {
    /**
     * Members
     */
    private _active: boolean;
    protected _spriteX: number;
    protected _spriteY: number;

    /**
     * Design
     */
    protected readonly _zoomStep: number = 0.05;

    /**
     * Constructor
     * @param element
     * @param active
     */
    public constructor(element: TElement, active: boolean = true) {
        // Parent constructor
        super(element);

        // Members
        this._active = active;
    }

    //<editor-fold desc="Public functions">
    /**
     * Reset
     */
    public reset(): void {
        this._spriteX = 0;
        this._spriteY = 0;
    }

    /**
     * Get graph coordinates
     * @param x
     * @param y
     * @private
     */
    public getGraphCoordinates(x: number, y: number) {
        this._spriteX = x;
        this._spriteY = y;
        return PIXI.InteractionData.prototype.getLocalPosition.call({global: { x: this._spriteX, y: this._spriteY}}, this.element);
    }
    //</editor-fold>

    //<editor-fold desc="Interactions">
    /**
     * On zoom
     * @param event
     */
    public onZoom(event: WheelEvent): void {
        if (!this._active) {
            return;
        }

        const direction = (event.deltaY < 0) ? 1 : -1;
        const factor = (1 + direction * this._zoomStep);

        this.emit("onZoom", event, event.clientX, event.clientY, factor);
    }
    //</editor-fold>

    //<editor-fold desc="Getter and Setter">
    public get active(): boolean {
        return this._active;
    }

    public set active(value: boolean) {
        this._active = value;
    }
    //</editor-fold>
}